/////////////////////////////////////////////////////////////////////////////
// Name:        alsainterface.cpp
// Purpose:     Audio mixer functions implementation via ALSA library.
// Author:      Yann Hamiaux
// Modified by: -
// Created:     08/15/2008
// Copyright:   Written for Guillemot Corporation by Yann Hamiaux
// Licence:     The GNU General Public License.
/////////////////////////////////////////////////////////////////////////////

#include <ctype.h>
#include <alsa/asoundlib.h>
#include "alsainterface.h"
#include "dmanager.h"


CAlsaInterface::CAlsaInterface( void )
{
}

CAlsaInterface::~CAlsaInterface( )
{
}

// ---------------------------------------------------------------------------
// Function: Callback_Mixer_Elem
//
// Callback when there is a volume/mute change.
// ---------------------------------------------------------------------------
int CAlsaInterface::Callback_Mixer_Elem(snd_mixer_elem_t *elem, unsigned int mask)
{
	const snd_mixer_selem_channel_id_t MIXER_CHANNELS[] = {
		SND_MIXER_SCHN_FRONT_LEFT, SND_MIXER_SCHN_FRONT_RIGHT,
		SND_MIXER_SCHN_REAR_LEFT, SND_MIXER_SCHN_REAR_RIGHT,
		SND_MIXER_SCHN_FRONT_CENTER, SND_MIXER_SCHN_WOOFER
		};
	long lValue, lMin, lMax;
	int  err, i, j, nValue, nMute, nChannelCount, *pnVolume;
	bool bRefresh, bMute, *pbMute;

	for(i=0;i<MAX_DEVICES;++i)
	{
		// We process only the device identified by the change!
		if (g_DEVICE_INFO[i].pALSA_elem_Volume==elem)
		{
			// Refresh all volumes and mute state for the given device.
			lMin = g_DEVICE_INFO[i].nVolMin;
			lMax = g_DEVICE_INFO[i].nVolMax;
			bRefresh = false;

			// --- Base on device ID, set variables to the parameters required ---
			switch(g_DEVICE_INFO[i].ulProductCode)
			{
			case DJCONSOLE_PRODUCT_CODE:
				nChannelCount = 6;
				pnVolume = &g_DEVICE_INFO[i].djConsole.nVolume[0];
				pbMute = &g_DEVICE_INFO[i].djConsole.bMute[0];
				break;
			case DJCONSOLE2_PRODUCT_CODE:
				nChannelCount = 4;
				pnVolume = &g_DEVICE_INFO[i].djConsoleMk2.nVolume[0];
				pbMute = &g_DEVICE_INFO[i].djConsoleMk2.bMute[0];
				break;
			case DJCONSOLERMX_PRODUCT_CODE:
				nChannelCount = 4;
				pnVolume = &g_DEVICE_INFO[i].djConsoleRmx.nVolume[0];
				pbMute = &g_DEVICE_INFO[i].djConsoleRmx.bMute[0];
				break;
			default:
				DEBUG_PRINT("ALSA Callback - Invalid product code.");
				return 0;	// Not valid.
			}

			// --- For each channel, query the new volume ---
			for(j=0;j<nChannelCount;++j)
			{
				err = snd_mixer_selem_get_playback_volume(elem, MIXER_CHANNELS[j], &lValue);
				nValue = 255 * (lValue-lMin) / (lMax-lMin);
				if (nValue!=pnVolume[j])
				{
					pnVolume[j] = nValue;
					bRefresh = true;
				}
			}

			// --- check: Mute channel 1/2 ---
			err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_FRONT_RIGHT, &nMute);
			bMute = (nMute==0);
			err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_FRONT_LEFT, &nMute);
			bMute&= (nMute==0);

			if (bMute!=pbMute[0])
			{
				pbMute[0] = bMute;
				bRefresh = true;
			}

			// Mute channel 3/4 : Hardware Mute Only for monitoring.

			// --- Mute channel 5/6 ---
			if (nChannelCount>=6)
			{
				err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_FRONT_CENTER, &nMute);
				bMute = (nMute==0);
				err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_WOOFER, &nMute);
				bMute&= (nMute==0);

				if (bMute!=pbMute[2])
				{
					pbMute[2] = bMute;
					bRefresh = true;

				}
			}

			g_DEVICE_INFO[i].bALSA_Refresh |= bRefresh;
			return 0;
		}
	}

	// We do not expect to receive a callback for a device which we
	// do not register the callback for.
	DEBUG_PRINT("ALSA Callback - Could not found device!?");
	return 0;
}

// ---------------------------------------------------------------------------
// Function: Callback_Mixer
//
// Not used (empty).
// ---------------------------------------------------------------------------
int CAlsaInterface::Callback_Mixer(snd_mixer_t *mixer, unsigned int mask, snd_mixer_elem_t *elem)
{
	return 0;
}

// ---------------------------------------------------------------------------
// Function: UpdateEventCallback
//
// Handle pending mixer events invoking callbacks (ALSA).
// ---------------------------------------------------------------------------
void CAlsaInterface::UpdateEventCallback( DJ_DEVICE_STATE *pState )
{
	snd_mixer_handle_events(pState->pALSA_Mixer);
}

// ---------------------------------------------------------------------------
// Function: OpenAlsa
//
// Open first ALSA mixer to match physical location and has a MixerName
// equal to 'USB Mixer'.
// ---------------------------------------------------------------------------
bool CAlsaInterface::Open( DJ_DEVICE_STATE *pState )
{
	char card[260];
	const char MIXER_NAME[] = "USB Mixer";
	const char *pLongName;
	const char *pMixerName;
	snd_ctl_t *ctl_handle;
	snd_ctl_card_info_t *hw_info = NULL;
	int result, err;
	int i, j, k;
	bool bFound;

	// Validate Arguments.
	if (pState==NULL)
		return false;

	i = 0;
	do
	{
		sprintf(card, "hw:%d", i);
		result = snd_ctl_open (&ctl_handle, card, 0);
		if (result==0)
		{
			snd_ctl_card_info_alloca(&hw_info);
			err = snd_ctl_card_info (ctl_handle, hw_info);
			pLongName = snd_ctl_card_info_get_longname(hw_info);
			pMixerName = snd_ctl_card_info_get_mixername (hw_info);
			bFound = false;
			for(j=0;!bFound && j<(int) (strlen(pLongName)-strlen(pState->szLocationID));++j)
			{
				bFound = true;
				for(k=0;bFound && k<(int) strlen(pState->szLocationID);++k)
					bFound = pLongName[j+k]==pState->szLocationID[k];
				for(k=0;bFound && k<(int) strlen(MIXER_NAME);++k)
					bFound = toupper(MIXER_NAME[k])==toupper(pMixerName[k]);
			}
			if (bFound)
			{
				snd_mixer_open(&pState->pALSA_Mixer, 0);
				err = snd_mixer_attach(pState->pALSA_Mixer, card);
				if (err==0)
				{
					err = snd_mixer_selem_register(pState->pALSA_Mixer, NULL, NULL);
					err = snd_mixer_load (pState->pALSA_Mixer);
					snd_mixer_set_callback(pState->pALSA_Mixer, Callback_Mixer);
					snd_mixer_set_callback_private(pState->pALSA_Mixer, NULL);
					pState->pALSA_SndCtl = ctl_handle;
					DEBUG_PRINT("ALSA - Open (%s)", pLongName);
					return true;
				}
				else
				{
					snd_mixer_close(pState->pALSA_Mixer);
					pState->pALSA_Mixer = NULL;
				}
			}
			snd_ctl_close(ctl_handle);
		}
		++i;
	} while(i<16 || result==0); 	// It could happen that device 0 will fail while our device is located at index 1.
					// Unplug-Replug the device could cause this issue to happen.
	DEBUG_PRINT("ALSA - Open (failed)");
	return false;
}

// ---------------------------------------------------------------------------
// Function: Close
//
// Close interfaces and set poiters to NULL.
// ---------------------------------------------------------------------------
bool CAlsaInterface::Close( DJ_DEVICE_STATE *pState )
{
	if (pState==NULL)
		return false;

	if (pState->bALSA_Init)
	{
		DEBUG_PRINT("ALSA - Close (%s)", pState->szLocationID);
		if (pState->pALSA_SndCtl!=0)
		{
			snd_ctl_close(pState->pALSA_SndCtl);
			pState->pALSA_SndCtl = 0;
		}
		if (pState->pALSA_Mixer!=0)
		{
			snd_mixer_close(pState->pALSA_Mixer);
			pState->pALSA_Mixer = 0;
		}
		pState->pALSA_SndCtl = NULL;
		pState->pALSA_Mixer = NULL;
		pState->pALSA_elem_Volume = NULL;
		pState->bALSA_Init = false;
	}

	return true;
}

// ---------------------------------------------------------------------------
// Function: RetrieveState
//
// Query ALSA mixer state base on the sound card.
// ---------------------------------------------------------------------------
bool CAlsaInterface::RetrieveState( DJ_DEVICE_STATE *pState )
{
	// --- Verify input parameter ---
	if (pState==NULL)
		return false;

	// --- Base on the product code, Retrieve Bulk State ---
	switch (pState->ulProductCode)
	{
	case DJCONSOLE_PRODUCT_CODE:
		return RetrieveState_DJConsole(pState);
	case DJCONSOLE2_PRODUCT_CODE:
		return RetrieveState_DJConsoleMk2(pState);
	case DJCONSOLERMX_PRODUCT_CODE:
		return RetrieveState_DJConsoleRmx(pState);
	}

	return false;
}

// ---------------------------------------------------------------------------
// Function: RetrieveState
//
// Retrieve ALSA state for DJ Console.
// ---------------------------------------------------------------------------
bool CAlsaInterface::RetrieveState_DJConsole( DJ_DEVICE_STATE *pState )
{
	snd_mixer_elem_t *elem;
	long lValue, lMinVal, lMaxVal;
	int nValue;
	int nElemIndex, nPlayback, err;

	DEBUG_PRINT("ALSA - Retrieve State (DJ Console)");
	elem = snd_mixer_first_elem (pState->pALSA_Mixer);
	nElemIndex = 0;
	while(elem!=NULL)
	{
		nPlayback = snd_mixer_selem_has_playback_volume(elem);
		snd_mixer_selem_get_playback_volume_range(elem, &lMinVal, &lMaxVal);
		pState->nVolMin = (int) lMinVal;
		pState->nVolMax = (int) lMaxVal;
		if (nPlayback!=0)
		{
			pState->pALSA_elem_Volume = elem;
			snd_mixer_elem_set_callback(elem, Callback_Mixer_Elem);
			snd_mixer_elem_set_callback_private(elem, NULL);

			// If at least one channel is unmute, it's consider unmute.
			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_FRONT_LEFT, &lValue);
			pState->djConsole.nVolume[0] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_FRONT_LEFT, &nValue);
			pState->djConsole.bMute[0] = (nValue==0);

			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_FRONT_RIGHT, &lValue);
			pState->djConsole.nVolume[1] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_FRONT_RIGHT, &nValue);
			pState->djConsole.bMute[0] &= (nValue==0);

			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_REAR_LEFT, &lValue);
			pState->djConsole.nVolume[2] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			// We take only the Hardware Mute 3/4 into account.
			// err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_REAR_LEFT, &nValue);

			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_REAR_RIGHT, &lValue);
			pState->djConsole.nVolume[3] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			// We take only the Hardware Mute 3/4 into account.
			// err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_REAR_RIGHT, &nValue);

			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_FRONT_CENTER, &lValue);
			pState->djConsole.nVolume[4] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_FRONT_CENTER, &nValue);
			pState->djConsole.bMute[2] = (nValue==0);


			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_WOOFER, &lValue);
			pState->djConsole.nVolume[5] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_WOOFER, &nValue);
			pState->djConsole.bMute[2] &= (nValue==0);
			return true;
		}

		elem = snd_mixer_elem_next(elem);
		++nElemIndex;
	};

	return false;
}

// ---------------------------------------------------------------------------
// Function: RetrieveState
//
// Retrieve ALSA state for DJ Console Mk2.
// ---------------------------------------------------------------------------
bool CAlsaInterface::RetrieveState_DJConsoleMk2( DJ_DEVICE_STATE *pState )
{
	snd_mixer_elem_t *elem;
	long lValue, lMinVal, lMaxVal;
	int nValue;
	int nElemIndex, nPlayback, err;

	DEBUG_PRINT("ALSA - Retrieve State (DJ Console Mk2)");
	elem = snd_mixer_first_elem (pState->pALSA_Mixer);
	nElemIndex = 0;
	while(elem!=NULL)
	{
		nPlayback = snd_mixer_selem_has_playback_volume(elem);
		snd_mixer_selem_get_playback_volume_range(elem, &lMinVal, &lMaxVal);
		pState->nVolMin = (int) lMinVal;
		pState->nVolMax = (int) lMaxVal;

		if (nPlayback!=0)
		{
			pState->pALSA_elem_Volume = elem;
			snd_mixer_elem_set_callback(elem, Callback_Mixer_Elem);
			snd_mixer_elem_set_callback_private(elem, NULL);

			// If at least one channel is unmute, it's consider unmute.
			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_FRONT_LEFT, &lValue);
			pState->djConsoleMk2.nVolume[0] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_FRONT_LEFT, &nValue);
			pState->djConsoleMk2.bMute[0] = (nValue==0);

			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_FRONT_RIGHT, &lValue);
			pState->djConsoleMk2.nVolume[1] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_FRONT_RIGHT, &nValue);
			pState->djConsoleMk2.bMute[0] &= (nValue==0);

			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_REAR_LEFT, &lValue);
			pState->djConsoleMk2.nVolume[2] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			// We take only the Hardware Mute 3/4 into account.
			// err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_REAR_LEFT, &nValue);

			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_REAR_RIGHT, &lValue);
			pState->djConsoleMk2.nVolume[3] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			// We take only the Hardware Mute 3/4 into account.
			// err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_REAR_RIGHT, &nValue);
			return true;
		}

		elem = snd_mixer_elem_next(elem);
		++nElemIndex;
	};

	return false;
}

// ---------------------------------------------------------------------------
// Function: RetrieveState
//
// Retrieve ALSA state for DJ Console Rmx.
// ---------------------------------------------------------------------------
bool CAlsaInterface::RetrieveState_DJConsoleRmx( DJ_DEVICE_STATE *pState )
{
	snd_mixer_elem_t *elem;
	long lValue, lMinVal, lMaxVal;
	int nValue;
	int nElemIndex, nPlayback, err;

	DEBUG_PRINT("ALSA - Retrieve State (DJ Console Rmx)");
	elem = snd_mixer_first_elem (pState->pALSA_Mixer);
	nElemIndex = 0;
	while(elem!=NULL)
	{
		nPlayback = snd_mixer_selem_has_playback_volume(elem);
		snd_mixer_selem_get_playback_volume_range(elem, &lMinVal, &lMaxVal);
		pState->nVolMin = (int) lMinVal;
		pState->nVolMax = (int) lMaxVal;

		if (nPlayback!=0)
		{
			pState->pALSA_elem_Volume = elem;
			snd_mixer_elem_set_callback(elem, Callback_Mixer_Elem);
			snd_mixer_elem_set_callback_private(elem, NULL);

			// If at least one channel is unmute, it's consider unmute.
			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_FRONT_LEFT, &lValue);
			pState->djConsoleRmx.nVolume[0] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_FRONT_LEFT, &nValue);
			pState->djConsoleRmx.bMute[0] = (nValue==0);

			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_FRONT_RIGHT, &lValue);
			pState->djConsoleRmx.nVolume[1] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_FRONT_RIGHT, &nValue);
			pState->djConsoleRmx.bMute[0] &= (nValue==0);

			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_REAR_LEFT, &lValue);
			pState->djConsoleRmx.nVolume[2] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			// We take only the Hardware Mute 3/4 into account.
			// err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_REAR_LEFT, &nValue);

			err = snd_mixer_selem_get_playback_volume(elem, SND_MIXER_SCHN_REAR_RIGHT, &lValue);
			pState->djConsoleRmx.nVolume[3] = 255 * (lValue-lMinVal) / (lMaxVal-lMinVal);
			// We take only the Hardware Mute 3/4 into account.
			// err = snd_mixer_selem_get_playback_switch(elem, SND_MIXER_SCHN_REAR_RIGHT, &nValue);

			return true;
		}

		elem = snd_mixer_elem_next(elem);
		++nElemIndex;
	};

	return false;
}


// ---------------------------------------------------------------------------
// Function: SetMute
//
// Set mute on a pair of audio channel (mute 3/4 is also done via HDJBulk).
// ---------------------------------------------------------------------------
bool CAlsaInterface::SetMute( DJ_DEVICE_STATE *pState, int nStereoChannel, bool bMute )
{
	snd_mixer_elem_t *elem;
	int nElemIndex, nPlayback, err, err2;
	long lMinVal, lMaxVal;
	int nValue;

	// --- Validate input parameters ---
	if (pState==NULL || nStereoChannel<0)
		return false; // Wrong interface.

	if (pState->ulProductCode==DJCONSOLE_PRODUCT_CODE)
	{
		if (nStereoChannel>2)
			return false;
	}
	else if (pState->ulProductCode==DJCONSOLE2_PRODUCT_CODE ||
		pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
	{
		if (nStereoChannel>1)
			return false;
	}
	else
		return false;

	elem = snd_mixer_first_elem (pState->pALSA_Mixer);
	nElemIndex = 0;
	nValue = (bMute) ? 0 : 1;

	while(elem!=NULL)
	{
		nPlayback = snd_mixer_selem_has_playback_volume(elem);
		snd_mixer_selem_get_playback_volume_range(elem, &lMinVal, &lMaxVal);
		if (nPlayback!=0)
		{
			switch(nStereoChannel)
			{
			case 0:
				err = snd_mixer_selem_set_playback_switch(elem, SND_MIXER_SCHN_FRONT_LEFT, nValue);
				err2= snd_mixer_selem_set_playback_switch(elem, SND_MIXER_SCHN_FRONT_RIGHT, nValue);
				break;
			case 1:
				err = snd_mixer_selem_set_playback_switch(elem, SND_MIXER_SCHN_REAR_LEFT, nValue);
				err2= snd_mixer_selem_set_playback_switch(elem, SND_MIXER_SCHN_REAR_RIGHT, nValue);
				break;
			case 2:
				err = snd_mixer_selem_set_playback_switch(elem, SND_MIXER_SCHN_FRONT_CENTER, nValue);
				err2= snd_mixer_selem_set_playback_switch(elem, SND_MIXER_SCHN_WOOFER, nValue);
				break;
			}
			goto finalize_set_mute;
		}
		elem = snd_mixer_elem_next(elem);
		++nElemIndex;
	}
	return false;


finalize_set_mute:
	if (pState->ulProductCode==DJCONSOLE_PRODUCT_CODE)
		pState->djConsole.bMute[nStereoChannel] = bMute;
	else if (pState->ulProductCode==DJCONSOLE2_PRODUCT_CODE)
		pState->djConsoleMk2.bMute[nStereoChannel] = bMute;
	else if (pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
		pState->djConsoleRmx.bMute[nStereoChannel] = bMute;

	return true;
}

// ---------------------------------------------------------------------------
// Function: SetVolume
//
// Set volume for a specific audio channel.
// ---------------------------------------------------------------------------
bool CAlsaInterface::SetVolume( DJ_DEVICE_STATE *pState, int nChannel, unsigned short usVolume )
{
	const _snd_mixer_selem_channel_id CHANNEL_TABLE[] = {
		SND_MIXER_SCHN_FRONT_LEFT, SND_MIXER_SCHN_FRONT_RIGHT,
		SND_MIXER_SCHN_REAR_LEFT, SND_MIXER_SCHN_REAR_RIGHT,
		SND_MIXER_SCHN_FRONT_CENTER, SND_MIXER_SCHN_WOOFER
		};
	snd_mixer_elem_t *elem;
	int nElemIndex, nPlayback, err;
	long lMinVal, lMaxVal;
	int nValue;

	// --- Validate input parameters ---
	if (pState==NULL || nChannel<0)
		return false; // Wrong interface.

	if (pState->ulProductCode==DJCONSOLE_PRODUCT_CODE)
	{
		if (nChannel>5)
			return false;
	}
	else if (pState->ulProductCode==DJCONSOLE2_PRODUCT_CODE ||
		pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
	{
		if (nChannel>3)
			return false;
	}
	else
		return false;

	elem = snd_mixer_first_elem (pState->pALSA_Mixer);
	nElemIndex = 0;
	while(elem!=NULL)
	{
		nPlayback = snd_mixer_selem_has_playback_volume(elem);
		snd_mixer_selem_get_playback_volume_range(elem, &lMinVal, &lMaxVal);
		nValue = (lMaxVal - lMinVal) * ((int) usVolume) / 255;
		if (nPlayback!=0)
		{
			err = snd_mixer_selem_set_playback_volume(elem, CHANNEL_TABLE[nChannel], nValue);
			goto finalize_set_mute;
		}
		elem = snd_mixer_elem_next(elem);
		++nElemIndex;
	}
	return false;


finalize_set_mute:
	if (pState->ulProductCode==DJCONSOLE_PRODUCT_CODE)
		pState->djConsole.nVolume[nChannel] = usVolume;
	else if (pState->ulProductCode==DJCONSOLE2_PRODUCT_CODE)
		pState->djConsoleMk2.nVolume[nChannel] = usVolume;
	else if (pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
		pState->djConsoleRmx.nVolume[nChannel] = usVolume;

	return true;
}


