/////////////////////////////////////////////////////////////////////////////
// Name:        bulkinterface.cpp
// Purpose:     Access the custom Hercules Driver which interface to
//		Hercules DJ products. The HID raw data is read/write via
//		the same custom driver.
//
//		Recently, some changes has been made in the Linux
//		kernel (HIDRAW component) to provide a way to retrieve
//		the physical location (which is required to bind multiple
//		driver (HID, ALSA, our custom driver) together when there
//		is multiple identical (same VID, PID, REV) device found.
// Author:      Yann Hamiaux
// Modified by: -
// Created:     03/19/2008
// Copyright:   Written for Guillemot Corporation by Yann Hamiaux
// Licence:     The GNU General Public License.
/////////////////////////////////////////////////////////////////////////////

#include <sys/ioctl.h>
#include <linux/types.h>
#include <sys/stat.h>

#include "bulkinterface.h"
#include "djdevioctls.h"


CBulkInterface::CBulkInterface( void )
{
	m_nBulkHandle = -1;
}

CBulkInterface::~CBulkInterface( )
{
}

// ---------------------------------------------------------------------------
// Function: BulkIoctl
//
// Process IOCTL for Bulk Interface. Please set m_nBulkHandle before any
// calls and reset it to -1 after. This allow many call to be done in a
// function without having to pass again and again the same parameter.
// ---------------------------------------------------------------------------
#define BULK_IOCTL_TIMEOUT 	1000	// Retry time-out is set to 1 second (1000 ms)
#define BULK_IOCTL_RETRY_DELAY	50	// Delay between retry is set to 50 ms.
int CBulkInterface::BulkIoctl( int nRequest, void *pParam )
{
	int nRetCode, nRetry = BULK_IOCTL_TIMEOUT / BULK_IOCTL_RETRY_DELAY;
	bool bPrintError = true;

	// --- m_nBulkHandle must be set ---
	if (m_nBulkHandle<0)
	{
		DEBUG_PRINT("Invalid bulk handle with nRequest=%d.", nRequest);
		return -1;
	}

	// ioctls excluded from retry.
	if ((unsigned int) nRequest==DJ_IOCTL_IS_DEVICE_ALIVE || 
		(unsigned int) nRequest==DJ_IOCTL_GET_CONTROL_DATA_INPUT_PACKET_SIZE ||
		(unsigned int) nRequest==DJ_IOCTL_GET_CONTROL_DATA_OUTPUT_PACKET_SIZE ||
		(unsigned int) nRequest==DJ_IOCTL_UNREGISTER_FOR_NETLINK_DEVICE_NOTIFICATIONS)
	{
		nRetry = 1;
		bPrintError = false;
	}

	while(nRetry-->0)
	{
		nRetCode = ioctl(m_nBulkHandle, nRequest, pParam);
		if (nRetCode>=0)
			return nRetCode;
		usleep(BULK_IOCTL_RETRY_DELAY * 1000);
	}
	
	if (bPrintError) {
		DEBUG_PRINT("BULK - IOCTL fail with code %d (nRequest=0x%.8X, errno=%d)", nRetCode, nRequest, errno);
	}

	return nRetCode;
}

// ---------------------------------------------------------------------------
// Function: SingleBulkIoctl
//
// Process single Bulk ioCtl which required nBulkHandle as input param.
// ---------------------------------------------------------------------------
int CBulkInterface::SingleBulkIoctl( int nBulkHandle, int nRequest, void *pParam )
{
	int nRetCode;

	m_nBulkHandle = nBulkHandle;
	nRetCode = BulkIoctl(nRequest, pParam);
	m_nBulkHandle = -1;
	return nRetCode;
}

// ---------------------------------------------------------------------------
// Function: OpenBulk
//
// Open /dev/hdjbulk<X> (where <x> equal nDeviceIndex) if possible.
// retrieve product code and physical location.
// ---------------------------------------------------------------------------
bool CBulkInterface::Open( int nDeviceIndex, DJ_DEVICE_STATE *pState )
{
	char		szBulkFilename[40];
	struct stat 	stat_buf;

	// Validate Arguments.
	if (pState==NULL)
		return false;

	// Initialize parameters.
	pState->bBULK_Init = false;
	pState->ulBULK_Version = 0x00000000;              // HDJBulk File Version (BCD).
	pState->ulBULK_ReadSize = 0;
	pState->ulBULK_OutSize = 0;
	pState->nBULK_FileIndex = -1;

	// Open Bulk interface.
	sprintf(szBulkFilename, "/dev/hdjbulk%d", nDeviceIndex);
	if (stat(szBulkFilename, &stat_buf)<0)
		return true;	// No error, but device is not present.

	pState->nBULK_FHandle = open(szBulkFilename, O_RDWR/*|O_NONBLOCK*/, 0);
	if (pState->nBULK_FHandle<0)
		return true;	// No error, but device is not present.

	// Read Product Code and Location Id.
	m_nBulkHandle = pState->nBULK_FHandle;
	BulkIoctl(DJ_IOCTL_GET_PRODUCT_CODE, &pState->ulProductCode);
	BulkIoctl(DJ_IOCTL_GET_LOCATION_ID, pState->szLocationID);
	BulkIoctl(DJ_IOCTL_GET_VERSION, &pState->ulBULK_Version);
	BulkIoctl(DJ_IOCTL_GET_CONTROL_DATA_INPUT_PACKET_SIZE, &pState->ulBULK_ReadSize);
	BulkIoctl(DJ_IOCTL_GET_CONTROL_DATA_OUTPUT_PACKET_SIZE, &pState->ulBULK_OutSize);
	BulkIoctl(DJ_IOCTL_ACQUIRE_NETLINK_UNIT, &pState->nNetlinkUnit);
	m_nBulkHandle = -1;

	// Validate parameters.
	if (pState->ulProductCode==DJCONSOLE_PRODUCT_UNKNOWN || 
		pState->ulProductCode>DJCONTROLSTEEL_PRODUCT_CODE ||
		pState->szLocationID[0]=='\0')
		goto init_bulk_fail;

	pState->bBULK_Init = true;
	pState->nBULK_FileIndex = nDeviceIndex;
	DEBUG_PRINT("BULK - Open (%s)", pState->szLocationID);
	return true;

init_bulk_fail:
	close(pState->nBULK_FHandle);
	DEBUG_PRINT("BULK - Open (failed)");
	return false;
}

// ---------------------------------------------------------------------------
// Function: Close
//
// Close and reset variables related with Bulk driver.
// ---------------------------------------------------------------------------
bool CBulkInterface::Close(DJ_DEVICE_STATE *pState )
{
	if (pState->bBULK_Init)
	{
		DEBUG_PRINT("BULK - Close (%s)", pState->szLocationID);
		close(pState->nBULK_FHandle);
		pState->bBULK_Init = false;
		pState->ulProductCode = DJCONSOLE_PRODUCT_UNKNOWN;// Not a valid product.
		pState->ulBULK_Version = 0x00000000;              // HDJBulk File Version (BCD).
		pState->nBULK_FileIndex = -1;
	}

	return true;
}

// ---------------------------------------------------------------------------
// Function: RetrieveBulkState
//
// Base on product code, retrieve bulk state by calling the proper function.
// ---------------------------------------------------------------------------
bool CBulkInterface::RetrieveState( DJ_DEVICE_STATE *pState )
{
	// --- Verify input parameter ---
	if (pState==NULL)
		return false;

	// --- Base on the product code, Retrieve Bulk State ---
	switch (pState->ulProductCode)
	{
	case DJCONSOLE_PRODUCT_CODE:
		return RetrieveState_DJConsole(pState);
	case DJCONSOLE2_PRODUCT_CODE:
		return RetrieveState_DJConsoleMk2(pState);
	case DJCONTROLLER_PRODUCT_CODE:
		return RetrieveState_DJControlMP3(pState);
	case DJCONSOLERMX_PRODUCT_CODE:
		return RetrieveState_DJConsoleRmx(pState);
	case DJCONTROLSTEEL_PRODUCT_CODE:
		return RetrieveState_DJControlSteel(pState);
	}

	return false;
}

// ---------------------------------------------------------------------------
// Function: RetrieveBulkState_DJConsole
//
// Retrieve bulk state for: DJ Console.
// ---------------------------------------------------------------------------
bool CBulkInterface::RetrieveState_DJConsole( DJ_DEVICE_STATE *pState )
{
	unsigned short usFirmwareVersion, usConfig, usMouseState, usSampleRate;
	unsigned short usTalkoverAtt;
	unsigned char ucTalkoverEnable;

	DJ_CONSOLE_STATE *pDJCState = &pState->djConsole;
	DEBUG_PRINT("BULK - Retrieve DJ Console state.");

	// --- Initialize parameters ---
	usFirmwareVersion = 0x0000;
	usConfig = 0;
	usMouseState = 0;
	usSampleRate = 48000;
	ucTalkoverEnable = 0;
	usTalkoverAtt = 0;

	// --- Retrieve Data from Device ---
	m_nBulkHandle = pState->nBULK_FHandle;
	BulkIoctl(DJ_IOCTL_GET_FIRMWARE_VERSION, &usFirmwareVersion);
	BulkIoctl(DJ_IOCTL_GET_DJCONSOLE_CONFIG, &usConfig);
	BulkIoctl(DJ_IOCTL_GET_SAMPLE_RATE, &usSampleRate);
	BulkIoctl(DJ_IOCTL_GET_MOUSE_STATE, &usMouseState);
	BulkIoctl(DJ_IOCTL_GET_TALKOVER_ATT, &usTalkoverAtt);
	BulkIoctl(DJ_IOCTL_GET_TALKOVER_ENABLE, &ucTalkoverEnable);
	m_nBulkHandle = -1;

	// --- Set Structure Parameters base on IOCTL ---
	// Set Firmware Version.
	pDJCState->lFirmwareVersion = (long) usFirmwareVersion;

	// Set Sample Rate.
	pDJCState->nSampleRate = (int) usSampleRate;

	// Set various properties base on Config WORD.
	pDJCState->bEnableDJMode = ((usConfig&DJC_AUDIOCFG_ADVANCED_MODE)!=0);
	pDJCState->bHeadphoneOnMaster = ((usConfig&DJC_AUDIOCFG_HEADPHONE_INPUT_SEL)!=0);

	pDJCState->bTalkoverAttenuation = (ucTalkoverEnable!=0);
	pDJCState->nTalkoverAttenuation = usTalkoverAtt;
	if (pDJCState->nTalkoverAttenuation<5)
		pDJCState->nTalkoverAttenuation = 3;
	else if (pDJCState->nTalkoverAttenuation<8)
		pDJCState->nTalkoverAttenuation = 6;
	else
		pDJCState->nTalkoverAttenuation = 9;

	pDJCState->bMute[1] = (usConfig&DJC_AUDIOCFG_MUTE_REAR_3_4)!=0;
	pDJCState->eMidiInput = (usConfig&DJC_AUDIOCFG_MIDI_IN_BUTTONS)!=0 ? DJ_CONSOLE_BUTTONS : MIDI_IN_PLUG;
	pDJCState->bEnableDJMouse = usMouseState!=0;

	if ((usConfig&DJC_AUDIOCFG_DIGITAL_IN)==0)
	{
		if (usConfig&DJC_AUDIOCFG_ANALOG_IN_LINE)
			pDJCState->eRecordingInput = LINE_IN;
		else
			pDJCState->eRecordingInput = MIC;
	}
	else
	{
		if (usConfig&DJC_AUDIOCFG_DIGITAL_IN_COAXIAL)
			pDJCState->eRecordingInput = DIGITAL_COAXIAL;
		else
			pDJCState->eRecordingInput = DIGITAL_OPTICAL;
	}

	m_nBulkHandle = -1;
	return true;
}

// ---------------------------------------------------------------------------
// Function: UpdateHIDSet_DJConsoleMk2
//
// Send HID data via the BULK driver using the current state.
// ---------------------------------------------------------------------------
bool CBulkInterface::UpdateHIDSet_DJConsoleMk2( DJ_DEVICE_STATE *pState )
{
	unsigned char *pBuf;

	// --- Set HID RAW throught Bulk Interface ---
	pBuf = (unsigned char *) malloc(pState->ulBULK_OutSize*2);
	memset(pBuf, 0, pState->ulBULK_OutSize*2);
	if (pState->djConsoleMk2.bLockAudioSource[0])
		pBuf[2] |= 0x10;
	if (pState->djConsoleMk2.bLockAudioSource[1])
		pBuf[2] |= 0x20;
	pBuf[5] |= 0x30;
	SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_OUTPUT_CONTROL_DATA, pBuf);
	
	free(pBuf);
	return true;
}

// ---------------------------------------------------------------------------
// Function: RetrieveBulkState_DJControlMP3
//
// Retrieve bulk state for: DJ Control MP3.
// ---------------------------------------------------------------------------
bool CBulkInterface::RetrieveState_DJControlMP3( DJ_DEVICE_STATE *pState )
{
	unsigned short usMouseState, usFirmwareVersion, usMidiChannel;

	DEBUG_PRINT("Retrieve DJ Control MP3 state. (bulk)");

	// --- Initialize parameters ---
	usMouseState = 0;
	usFirmwareVersion = 0;

	// --- Retrieve Data from Device ---
	m_nBulkHandle = pState->nBULK_FHandle;
	BulkIoctl(DJ_IOCTL_GET_MOUSE_STATE, &usMouseState);
	BulkIoctl(DJ_IOCTL_GET_FIRMWARE_VERSION, &usFirmwareVersion);
	BulkIoctl(DJ_IOCTL_GET_MIDI_CHANNEL, &usMidiChannel);
	m_nBulkHandle = -1;

	pState->djControlMP3.bEnableDJMouse = (usMouseState!=0);
	pState->djControlMP3.lFirmwareVersion = usFirmwareVersion;
	pState->djControlMP3.nMidiChannel = usMidiChannel;

	return true;
}

// ---------------------------------------------------------------------------
// Function: RetrieveBulkState_DJConsoleMk2
//
// Retrieve bulk state for: DJ Console Mk2.
// ---------------------------------------------------------------------------
bool CBulkInterface::RetrieveState_DJConsoleMk2( DJ_DEVICE_STATE *pState )
{
	unsigned short usMouseState, usFirmwareVersion;
	unsigned short usAudioCfg, usCrossFaderStyle, usCrossFaderLock;
	unsigned short usTalkoverAtt;
	unsigned char ucTalkoverEnable;
	DJ_CONSOLE_MK2_STATE *pMk2State = &pState->djConsoleMk2;

	DEBUG_PRINT("Retrieve DJ Console Mk2 state. (bulk)");

	// --- Initialize parameters ---
	usMouseState = 0;
	usFirmwareVersion = 0;
	usTalkoverAtt = 6; // -3 dB.
	usAudioCfg = 0;
	usCrossFaderStyle = 0;
	usCrossFaderLock = 0;
	ucTalkoverEnable = 0;

	// --- Retrieve Data from Device ---
	m_nBulkHandle = pState->nBULK_FHandle;
	BulkIoctl(DJ_IOCTL_GET_MOUSE_STATE, &usMouseState);
	BulkIoctl(DJ_IOCTL_GET_FIRMWARE_VERSION, &usFirmwareVersion);
	BulkIoctl(DJ_IOCTL_GET_TALKOVER_ATT, &usTalkoverAtt);
	BulkIoctl(DJ_IOCTL_GET_AUDIO_CONFIG, &usAudioCfg);
	BulkIoctl(DJ_IOCTL_GET_CROSSFADER_STYLE, &usCrossFaderStyle);
	BulkIoctl(DJ_IOCTL_GET_CFADER_LOCK, &usCrossFaderLock);
	BulkIoctl(DJ_IOCTL_GET_TALKOVER_ATT, &usTalkoverAtt);
	BulkIoctl(DJ_IOCTL_GET_TALKOVER_ENABLE, &ucTalkoverEnable);

	// --- Set Structure Parameters base on IOCTL ---
	// Set Firmware Version
	pMk2State->lFirmwareVersion = (long) usFirmwareVersion;
	pMk2State->lFirmwareVersion+= (long) ((usAudioCfg&0xFF00)<<8);

	// Set DJ Mouse.
	pMk2State->bEnableDJMouse = (usMouseState!=0);

	// Set Talkover Attenuation.
	pMk2State->bTalkoverAttenuation = (ucTalkoverEnable!=0);
	pMk2State->nTalkoverAttenuation = usTalkoverAtt;
	if (pMk2State->nTalkoverAttenuation<5)
		pMk2State->nTalkoverAttenuation = 3;
	else if (pMk2State->nTalkoverAttenuation<8)
		pMk2State->nTalkoverAttenuation = 6;
	else
		pMk2State->nTalkoverAttenuation = 9;

	// Set Hw-Mute-3/4, Use Ext. Mixer and Input Mic.
	pMk2State->bMute[1] = (usAudioCfg&DJMK2_AUDIOCFG_HW_MUTE_CH_3_4)!=0;
	pMk2State->bUseExternalMixer = (usAudioCfg&DJMK2_AUDIOCFG_USE_EXTERNAL_MIXER)!=0;
	pMk2State->bEnableInputMic = (usAudioCfg&DJMK2_AUDIOCFG_RECORD_INPUT_LINE_IN)==0;

	// Set Cross Fader Style base on curve and lock.
	if ((usCrossFaderStyle&0xFF)<13)
		pMk2State->eCrossFaderCurve = BEATMIX;
	else if ((usCrossFaderStyle&0xFF)<20)
		pMk2State->eCrossFaderCurve = DIPPED;
	else
		pMk2State->eCrossFaderCurve = SCRATCH;

	if ((usCrossFaderLock&0x30)==0x10)
		pMk2State->eCrossFaderCurve = LOCK_ON_A;
	else if ((usCrossFaderLock&0x30)==0x20)
		pMk2State->eCrossFaderCurve = LOCK_ON_B;
	else if ((usCrossFaderLock&0x30)==0x30)
		pMk2State->eCrossFaderCurve = LOCK_ON_CENTER;

	m_nBulkHandle = -1;
	return true;
}

// ---------------------------------------------------------------------------
// Function: RetrieveBulkState_DJConsoleRmx
//
// Retrieve bulk state for: DJ Console Rmx.
// ---------------------------------------------------------------------------
bool CBulkInterface::RetrieveState_DJConsoleRmx( DJ_DEVICE_STATE *pState )
{
	unsigned short usFirmwareVersion, usMidiChannel;
	unsigned short usAudioCfg, usJogWheelSensitivity, usJogWheelLock;
	unsigned short usTalkoverAtt;
	unsigned char  ucTalkoverEnable;
	snd_hdj_caps   hdjcaps;
	DJ_CONSOLE_RMX_STATE *pDJRmxState = &pState->djConsoleRmx;

	DEBUG_PRINT("Retrieve DJ Console Rmx state. (bulk)");

	// --- Initialize parameters ---
	usFirmwareVersion = 0x0000;
	usTalkoverAtt = 6;
	usMidiChannel = 0;
	usAudioCfg = 0;
	usJogWheelSensitivity = 0;
	usJogWheelLock = 0;
	ucTalkoverEnable = 0;
	memset(&hdjcaps, 0, sizeof(snd_hdj_caps));

	// --- Retrieve Data from Device ---
	m_nBulkHandle = pState->nBULK_FHandle;
	BulkIoctl(DJ_IOCTL_GET_FIRMWARE_VERSION, &usFirmwareVersion);
	BulkIoctl(DJ_IOCTL_GET_MIDI_CHANNEL, &usMidiChannel);
	BulkIoctl(DJ_IOCTL_GET_JOG_WHEEL_LOCK_SETTING, &usJogWheelLock);
	BulkIoctl(DJ_IOCTL_GET_JOG_WHEEL_SENSITIVITY, &usJogWheelSensitivity);
	BulkIoctl(DJ_IOCTL_GET_AUDIO_CONFIG, &usAudioCfg);
	BulkIoctl(DJ_IOCTL_GET_TALKOVER_ATT, &usTalkoverAtt);
	BulkIoctl(DJ_IOCTL_GET_TALKOVER_ENABLE, &ucTalkoverEnable);
	BulkIoctl(DJ_IOCTL_GET_DEVICE_CAPS, &hdjcaps);

	// --- Set Structure Parameters base on IOCTL ---
	// Set Firmware Version
	pDJRmxState->lFirmwareVersion = (long) ((usFirmwareVersion&0xF000)<<12);
	pDJRmxState->lFirmwareVersion|= ((hdjcaps.controller_board_version&0xFF)<<8);
	pDJRmxState->lFirmwareVersion|= hdjcaps.audio_board_version&0xFF;
	if (hdjcaps.controller_board_upgradeable)
		pDJRmxState->lFirmwareVersion |= 0x010000;

	// Set MIDI Channel
	pDJRmxState->nMidiChannel = (long) usMidiChannel;

	// Set Jog-Wheel Lock Setting.
	pDJRmxState->eJogWheelsLock = ENABLE;
	if (usJogWheelLock&COMMON_JOGWHEEL_LOCK_DECK_A) 
	{
		if (usJogWheelLock&COMMON_JOGWHEEL_LOCK_DECK_B)
			pDJRmxState->eJogWheelsLock = DISABLE_BOTH;
		else
			pDJRmxState->eJogWheelsLock = DISABLE_A;
	}
	else if (usJogWheelLock&COMMON_JOGWHEEL_LOCK_DECK_B)
		pDJRmxState->eJogWheelsLock = DISABLE_B;

	// Set Jog-Wheel Sensitivity.
	if (usJogWheelSensitivity==COMMON_JOGWHEEL_SENSITIVITY_LESS)
		pDJRmxState->eJogWheelsSensitivity = LESS;
	else if (usJogWheelSensitivity==COMMON_JOGWHEEL_SENSITIVITY_LEAST)
		pDJRmxState->eJogWheelsSensitivity = LEAST;
	else
		pDJRmxState->eJogWheelsSensitivity = NORMAL;

	// Set Talkover Attenuation
	pDJRmxState->bTalkoverAttenuation = (ucTalkoverEnable!=0);
	pDJRmxState->nTalkoverAttenuation = usTalkoverAtt;
	if (pDJRmxState->nTalkoverAttenuation<5)
		pDJRmxState->nTalkoverAttenuation = 3;
	else if (pDJRmxState->nTalkoverAttenuation<8)
		pDJRmxState->nTalkoverAttenuation = 6;
	else
		pDJRmxState->nTalkoverAttenuation = 9;

	// Set parameters base on audio config.
	pDJRmxState->bEnableInputMic = (usAudioCfg&DJRMX_AUDIOCFG_RECORD_INPUT_LINE_IN)==0;
	if (usAudioCfg&DJRMX_AUDIOCFG_MAIN_OUT_MIX)
	{
		if (usAudioCfg&DJRMX_AUDIOCFG_RECORD_INPUT_LINE_IN)
			pDJRmxState->eOutput12Mix = MIX_WITH_LINE_IN;
		else
			pDJRmxState->eOutput12Mix = MIX_WITH_MIC;
	}
	else
		pDJRmxState->eOutput12Mix = NONE;
	pDJRmxState->bMute[1] = (usAudioCfg&DJRMX_AUDIOCFG_MUTE_34)!=0;

	// Set Volume Linkage by default.
	pDJRmxState->bVolumeLink[0] = true;
	pDJRmxState->bVolumeLink[1] = true;

	m_nBulkHandle = -1;
	return true;
}

// ---------------------------------------------------------------------------
// Function: RetrieveBulkState_DJControlSteel
//
// Retrieve bulk state for: DJ Control Steel.
// ---------------------------------------------------------------------------
bool CBulkInterface::RetrieveState_DJControlSteel( DJ_DEVICE_STATE *pState )
{
	unsigned short usFirmwareVersion, usMidiChannel;
	unsigned short usJogWheelLock, usJogWheelSensitivity;
	DJ_CONTROL_STEEL_STATE *pSteelState = &pState->djControlSteel;

	DEBUG_PRINT("Retrieve DJ Control Steel state. (bulk)");

	// --- Initialize parameters ---
	usFirmwareVersion = 0x0000;
	usMidiChannel = 0;
	usJogWheelSensitivity = 0;
	usJogWheelLock = 0;

	// --- Retrieve Data from Device ---
	m_nBulkHandle = pState->nBULK_FHandle;
	BulkIoctl(DJ_IOCTL_GET_FIRMWARE_VERSION, &usFirmwareVersion);
	BulkIoctl(DJ_IOCTL_GET_MIDI_CHANNEL, &usMidiChannel);
	BulkIoctl(DJ_IOCTL_GET_JOG_WHEEL_LOCK_SETTING, &usJogWheelLock);
	BulkIoctl(DJ_IOCTL_GET_JOG_WHEEL_SENSITIVITY, &usJogWheelSensitivity);

	// --- Set Structure Parameters base on IOCTL ---
	// Get Firmware Version.
	pSteelState->lFirmwareVersion = (long) usFirmwareVersion;

	// Get MIDI Channel.
	pSteelState->nMidiChannel = (long) usMidiChannel;

	// Set Jog-Wheel Lock Setting.
	pSteelState->eJogWheelsLock = ENABLE;
	if (usJogWheelLock&COMMON_JOGWHEEL_LOCK_DECK_A) 
	{
		if (usJogWheelLock&COMMON_JOGWHEEL_LOCK_DECK_B)
			pSteelState->eJogWheelsLock = DISABLE_BOTH;
		else
			pSteelState->eJogWheelsLock = DISABLE_A;
	}
	else if (usJogWheelLock&COMMON_JOGWHEEL_LOCK_DECK_B)
		pSteelState->eJogWheelsLock = DISABLE_B;

	// Set Jog-Wheel Sensitivity.
	if (usJogWheelSensitivity==COMMON_JOGWHEEL_SENSITIVITY_LESS)
		pSteelState->eJogWheelsSensitivity = LESS;
	else if (usJogWheelSensitivity==COMMON_JOGWHEEL_SENSITIVITY_LEAST)
		pSteelState->eJogWheelsSensitivity = LEAST;
	else
		pSteelState->eJogWheelsSensitivity = NORMAL;
	
	m_nBulkHandle = -1;
	return true;
}

#define PROD_FLAGS_DJC		0x0001
#define PROD_FLAGS_MK2		0x0002
#define PROD_FLAGS_MP3		0x0004
#define PROD_FLAGS_RMX		0x0008
#define PROD_FLAGS_STEEL	0x0010
// ---------------------------------------------------------------------------
// Function: CheckIfSupported
//
// Allow to check if the device state match a support flag.
// This is use by Set functions on the BulkInterface in order to make sure
// the function is not call by another device.
// ---------------------------------------------------------------------------
bool CBulkInterface::CheckIfSupported(DJ_DEVICE_STATE *pState, int nSupportFlags)
{
	if (pState==NULL)
		return false;

	switch(pState->ulProductCode)
	{
	case DJCONSOLE_PRODUCT_CODE:
		return ((nSupportFlags&PROD_FLAGS_DJC)!=0);
	case DJCONSOLE2_PRODUCT_CODE:
		return ((nSupportFlags&PROD_FLAGS_MK2)!=0);
	case DJCONTROLLER_PRODUCT_CODE:
		return ((nSupportFlags&PROD_FLAGS_MP3)!=0);
	case DJCONSOLERMX_PRODUCT_CODE:
		return ((nSupportFlags&PROD_FLAGS_RMX)!=0);
	case DJCONTROLSTEEL_PRODUCT_CODE:
		return ((nSupportFlags&PROD_FLAGS_STEEL)!=0);
	}
	return false;
}

// ---------------------------------------------------------------------------
// Function: IsDeviceAlive
//
// Check if device is alive.
// ---------------------------------------------------------------------------
bool CBulkInterface::IsDeviceAlive( DJ_DEVICE_STATE *pState )
{
	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_IS_DEVICE_ALIVE, NULL)!=0)
		return false;

	return true;
}


// ---------------------------------------------------------------------------
// Function: EnableDJMode
//
// This feature is available only on DJ Console.
// ---------------------------------------------------------------------------
bool CBulkInterface::EnableDJMode( bool bEnable, DJ_DEVICE_STATE *pState )			// DJC
{
	unsigned long ulConfig;

	if (!CheckIfSupported(pState, PROD_FLAGS_DJC))
		return false;	// Check Argument -> Not supported.

	ulConfig = (bEnable) ? DJC_AUDIOCFG_ADVANCED_MODE : 0;
	ulConfig <<= 16;
	ulConfig |= DJC_AUDIOCFG_ADVANCED_MODE;

	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_DJCONSOLE_CONFIG, &ulConfig)<0)
		return false;

	pState->djConsole.bEnableDJMode = bEnable;
	DEBUG_PRINT("bulk-EnableDJMode (bEnable=%d) with success", bEnable);
	return true;
}

// ---------------------------------------------------------------------------
// Function: EnableDJMouse
//
// This feature is only available on DJ Console Mk2. Enabling DJ Mouse for
// DJ Console is done through HID Interface.
// ---------------------------------------------------------------------------
bool CBulkInterface::EnableDJMouse( bool bEnable, DJ_DEVICE_STATE *pState )			// DJC, Mk2, MP3
{
	if (!CheckIfSupported(pState, PROD_FLAGS_DJC|PROD_FLAGS_MK2|PROD_FLAGS_MP3))
		return false;	// Check Argument -> Not supported.
	if (bEnable)
	{
		if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_ENABLE_MOUSE, NULL)<0)
			return false;
	}
	else
	{
		if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_DISABLE_MOUSE, NULL)<0)
			return false;
	}

	if (pState->ulProductCode==DJCONSOLE_PRODUCT_CODE)
		pState->djConsole.bEnableDJMouse = bEnable;
	else if (pState->ulProductCode==DJCONSOLE2_PRODUCT_CODE)
		pState->djConsoleMk2.bEnableDJMouse = bEnable;
	else
		pState->djControlMP3.bEnableDJMouse = bEnable;

	DEBUG_PRINT("bulk-EnableDJMouse (bEnable=%d) with success", bEnable);
	return true;
}

// ---------------------------------------------------------------------------
// Function: EnableTalkover
//
// Enable/Disable Talkover Attenuation. For DJ Console Mk2 and Rmx, you need
// to set Talkover Attenuation to 0 if disable.
// ---------------------------------------------------------------------------
bool CBulkInterface::EnableTalkover( bool bEnable, DJ_DEVICE_STATE *pState )			// DJC, Mk2, Rmx
{
	unsigned char ucValue;

	if (!CheckIfSupported(pState, PROD_FLAGS_DJC|PROD_FLAGS_MK2|PROD_FLAGS_RMX))
		return false;	// Check Argument -> Not supported.

	// Enable/Disable Talkover on DJ Console.
	ucValue = bEnable ? 1 : 0;
	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_TALKOVER_ENABLE, &ucValue)<0)
		return false;

	if (pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
		pState->djConsoleRmx.bTalkoverAttenuation = bEnable;
	else if (pState->ulProductCode==DJCONSOLE_PRODUCT_CODE)
		pState->djConsole.bTalkoverAttenuation = bEnable;
	else if (pState->ulProductCode==DJCONSOLE2_PRODUCT_CODE)
		pState->djConsoleMk2.bTalkoverAttenuation = bEnable;

	DEBUG_PRINT("bulk-EnableTalkover (bEnable=%d) with success", bEnable);
	return true;
}

// ---------------------------------------------------------------------------
// Function: LockJogWheel
//
// Lock/Unlock JogWheel (DJ Console Mk2 has a jogwheel on PC which is
// API Software Lock).
// ---------------------------------------------------------------------------
bool CBulkInterface::LockJogWheel( JOG_WHEELS_DIS eJogWheelDisable, DJ_DEVICE_STATE *pState )	// Rmx, Steel
{
	unsigned short usValue;

	if (!CheckIfSupported(pState, PROD_FLAGS_RMX|PROD_FLAGS_STEEL))
		return false;	// Check Argument -> Not supported.

	
	if (pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE || pState->ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
	{
		// --- Set Jog Wheel Lock state for DJ Console Rmx ---
		switch(eJogWheelDisable)
		{
			case ENABLE:
				usValue = COMMON_JOGWHEEL_LOCK_NONE;
				break;
			case DISABLE_A:
				usValue = COMMON_JOGWHEEL_LOCK_DECK_A;
				break;
			case DISABLE_B:
				usValue = COMMON_JOGWHEEL_LOCK_DECK_B;
				break;
			case DISABLE_BOTH:
				usValue = COMMON_JOGWHEEL_LOCK_DECK_BOTH;
				break;
		}
		if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_JOG_WHEEL_LOCK_SETTING, &usValue)<0)
			return false;

		if (pState->ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
			pState->djControlSteel.eJogWheelsLock = eJogWheelDisable;
		else
			pState->djConsoleRmx.eJogWheelsLock = eJogWheelDisable;
	}

	DEBUG_PRINT("bulk-LockJogWheel (eJogWheelDisable=%d) with success", eJogWheelDisable);
	return true;
}

// ---------------------------------------------------------------------------
// Function: MuteChannel34
// ---------------------------------------------------------------------------
bool CBulkInterface::MuteChannel34( bool bMute34, DJ_DEVICE_STATE *pState )			// DJC, Mk2, Rmx
{
	unsigned long ulConfig;
	unsigned short usConfig;

	if (!CheckIfSupported(pState, PROD_FLAGS_DJC|PROD_FLAGS_MK2|PROD_FLAGS_RMX))
		return false;	// Check Argument -> Not supported.

	if (pState->ulProductCode==DJCONSOLE_PRODUCT_CODE)
	{
		// --- Set Mute3/4 on DJ Console ---
		ulConfig = 0;
		if (bMute34)
			ulConfig = DJC_AUDIOCFG_MUTE_REAR_3_4;
		ulConfig = (ulConfig<<16) | DJC_AUDIOCFG_MUTE_REAR_3_4;

		if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_DJCONSOLE_CONFIG, &ulConfig)<0)
			return false;
		pState->djConsole.bMute[1] = bMute34;
	}
	else if (pState->ulProductCode==DJCONSOLE2_PRODUCT_CODE)
	{
		// --- Set Mute3/4 on DJ Console Mk2 ---
		usConfig = 0;
		if (bMute34)
			usConfig = DJMK2_AUDIOCFG_HW_MUTE_CH_3_4;
		usConfig = (usConfig<<8) | DJMK2_AUDIOCFG_HW_MUTE_CH_3_4;

		if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_AUDIO_CONFIG, &usConfig)<0)
			return false;
		pState->djConsoleMk2.bMute[1] = bMute34;
	}
	else if (pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
	{
		// --- Set Mute3/4 on DJ Console Rmx ---
		usConfig = 0;
		if (bMute34)
			usConfig = DJRMX_AUDIOCFG_MUTE_34;
		usConfig = (usConfig<<8) | DJRMX_AUDIOCFG_MUTE_34;

		if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_AUDIO_CONFIG, &usConfig)<0)
			return false;
		pState->djConsoleRmx.bMute[1] = bMute34;
	}

	DEBUG_PRINT("bulk-MuteChannel34 (bMute34=%d) with success", bMute34);
	return true;
}

// ---------------------------------------------------------------------------
// Function: SetAudioSource
// ---------------------------------------------------------------------------
bool CBulkInterface::SetAudioSource( int nDeck, AUDIO_SOURCE eAudioSource, DJ_DEVICE_STATE *pState )// Mk2
{
	unsigned short usConfig;

	if (!CheckIfSupported(pState, PROD_FLAGS_MK2))
		return false;	// Check Argument -> Not supported.

	if (nDeck==0)
	{
		usConfig = (eAudioSource==0) ? DJMK2_AUDIOCFG_DECK_A_OUTPUT_1_2 : 0;
		usConfig = (usConfig<<8) | DJMK2_AUDIOCFG_DECK_A_OUTPUT_1_2;
	}
	else if (nDeck==1)
	{
		usConfig = (eAudioSource==0) ? DJMK2_AUDIOCFG_DECK_B_OUTPUT_3_4 : 0;
		usConfig = (usConfig<<8) | DJMK2_AUDIOCFG_DECK_B_OUTPUT_3_4;
	}
	else // Deck not supported.
		return false;

	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_AUDIO_CONFIG, &usConfig)<0)
		return false;

	pState->djConsoleMk2.eAudioSourceSelection[nDeck] = eAudioSource;
	DEBUG_PRINT("bulk-SetAudioSource (nDeck=%d, eAudioSource=%d) with success", nDeck, eAudioSource);
	return true;
}

// ---------------------------------------------------------------------------
// Function: SetCrossFaderCurve
// ---------------------------------------------------------------------------
bool CBulkInterface::SetCrossFaderCurve( CROSSFADER_SEL eCurve, DJ_DEVICE_STATE *pState )	// Mk2
{
	unsigned short usCFaderLock, usCFaderStyle;
	int nRetCode, nRetCode2;

	if (!CheckIfSupported(pState, PROD_FLAGS_MK2))
		return false;	// Check Argument -> Not supported.

	// --- Set Cross Fader Curve ---
	switch(eCurve)
	{
		case BEATMIX:
			usCFaderLock = 0;
			usCFaderStyle = 10|(64<<8);
			break;
		case DIPPED:
			usCFaderLock = 0;
			usCFaderStyle = 16|(128<<8);
			break;
		case SCRATCH:
			usCFaderLock = 0;
			usCFaderStyle = 56|(208<<8);
			break;
		case LOCK_ON_A:
			usCFaderLock = 1<<4;
			usCFaderStyle = 16|(128<<8);
			break;
		case LOCK_ON_B:
			usCFaderLock = 2<<4;
			usCFaderStyle = 16|(128<<8);
			break;
		case LOCK_ON_CENTER:
			usCFaderLock = 3<<4;
			usCFaderStyle = 16|(128<<8);
			break;
	}
	
	nRetCode = SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_CFADER_LOCK, &usCFaderLock);
	nRetCode2= SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_CROSSFADER_STYLE, &usCFaderStyle);
	if (nRetCode<0 || nRetCode2<0)
	{
		// --- If one fail, retrieve again the information in order to get the right cross fader style ---
		RetrieveState(pState);
		return false;
	}

	pState->djConsoleMk2.eCrossFaderCurve = eCurve;
	DEBUG_PRINT("bulk-SetCrossFaderCurve (eCurve=%d) with success", eCurve);
	return true;
}

// ---------------------------------------------------------------------------
// Function: SetHeadphoneToMainOut
// ---------------------------------------------------------------------------
bool CBulkInterface::SetHeadphoneToMainOut( bool bHeadphoneToMain, DJ_DEVICE_STATE *pState )	// DJC
{
	unsigned long ulConfig;

	if (!CheckIfSupported(pState, PROD_FLAGS_DJC))
		return false;	// Check Argument -> Not supported.

	ulConfig = (bHeadphoneToMain) ? DJC_AUDIOCFG_HEADPHONE_INPUT_SEL : 0;
	ulConfig = (ulConfig<<16) | DJC_AUDIOCFG_HEADPHONE_INPUT_SEL;

	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_DJCONSOLE_CONFIG, &ulConfig)<0)
		return false;

	pState->djConsole.bHeadphoneOnMaster = bHeadphoneToMain;
	DEBUG_PRINT("bulk-SetHeadphoneToMainOut (bHeadphoneToMain=%d) with success", bHeadphoneToMain);
	return true;
}

// ---------------------------------------------------------------------------
// Function: SetJogsSensitivity
// ---------------------------------------------------------------------------
bool CBulkInterface::SetJogsSensitivity( JOGS_SENSITIVE eSensitivity, DJ_DEVICE_STATE *pState )	// Rmx, Steel
{
	unsigned short usConfig;

	if (!CheckIfSupported(pState, PROD_FLAGS_RMX|PROD_FLAGS_STEEL))
		return false;	// Check Argument -> Not supported.

	if (eSensitivity==LESS)
		usConfig = COMMON_JOGWHEEL_SENSITIVITY_LESS;
	else if (eSensitivity==LEAST)
		usConfig = COMMON_JOGWHEEL_SENSITIVITY_LEAST;
	else
		usConfig = COMMON_JOGWHEEL_SENSITIVITY_NORMAL;

	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_JOG_WHEEL_SENSITIVITY, &usConfig)<0)
		return false;

	if (pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
		pState->djConsoleRmx.eJogWheelsSensitivity = eSensitivity;
	else
		pState->djControlSteel.eJogWheelsSensitivity = eSensitivity;

	DEBUG_PRINT("bulk-SetJogsSensitivity (eSensitivity=%d) with success", eSensitivity);
	return true;
}

// ---------------------------------------------------------------------------
// Function: SetMidiChannel
// ---------------------------------------------------------------------------
bool CBulkInterface::SetMidiChannel( int nMidiChannel, DJ_DEVICE_STATE *pState )		// Rmx, MP3, Steel
{
	unsigned short usValue;

	if (!CheckIfSupported(pState, PROD_FLAGS_RMX|PROD_FLAGS_STEEL|PROD_FLAGS_MP3))
		return false;	// Check Argument -> Not supported.

	usValue = nMidiChannel;
	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_MIDI_CHANNEL, &usValue)<0)
		return false;

	if (pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
		pState->djConsoleRmx.nMidiChannel = nMidiChannel;
	else if (pState->ulProductCode==DJCONTROLLER_PRODUCT_CODE)
		pState->djControlMP3.nMidiChannel = nMidiChannel;
	else
		pState->djControlSteel.nMidiChannel = nMidiChannel;

	DEBUG_PRINT("bulk-SetMidiChannel (nMidiChannel=%d) with success", nMidiChannel);
	return true;
}

// ---------------------------------------------------------------------------
// Function: SetMidiInput
// ---------------------------------------------------------------------------
bool CBulkInterface::SetMidiInput( MIDI_INPUT_SEL eMidiInput, DJ_DEVICE_STATE *pState )		// DJC
{
	unsigned long ulValue;

	if (!CheckIfSupported(pState, PROD_FLAGS_DJC))
		return false;	// Check Argument -> Not supported.

	ulValue = (eMidiInput==DJ_CONSOLE_BUTTONS) ? DJC_AUDIOCFG_MIDI_IN_BUTTONS : 0;
	ulValue = (ulValue<<16) | DJC_AUDIOCFG_MIDI_IN_BUTTONS;
	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_DJCONSOLE_CONFIG, &ulValue)<0)
		return false;

	pState->djConsole.eMidiInput = eMidiInput;
	DEBUG_PRINT("bulk-SetMidiInput (eMidiInput=%d) with success", eMidiInput);
	return true;
}

// ---------------------------------------------------------------------------
// Function: SetSampleRate
// ---------------------------------------------------------------------------
bool CBulkInterface::SetSampleRate( int nSampleRate, DJ_DEVICE_STATE *pState )			// DJC
{
	unsigned short usValue;

	if (!CheckIfSupported(pState, PROD_FLAGS_DJC))
		return false;	// Check Argument -> Not supported.

	if (nSampleRate!=44100 && nSampleRate!=48000)
		return false;

	// Set Audio Configuration.
	usValue = nSampleRate;
	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_SAMPLE_RATE, &usValue)<0)
		return false;

	// After this operation, the DJ Console need some time in order to depart/arrive with a new descriptor
	// specifying the sample rate.
	pState->djConsole.nSampleRate = nSampleRate;
	DEBUG_PRINT("bulk-SetSampleRate (nSampleRate=%d) with success", nSampleRate);
	return true;
}

// ---------------------------------------------------------------------------
// Function: SetRecordInput
// ---------------------------------------------------------------------------
bool CBulkInterface::SetRecordInput( RECORD_INPUT eRecordInput, DJ_DEVICE_STATE *pState )	// DJC, Mk2, Rmx
{
	unsigned long ulValue;
	unsigned short usValue;

	if (!CheckIfSupported(pState, PROD_FLAGS_DJC|PROD_FLAGS_MK2|PROD_FLAGS_RMX))
		return false;	// Check Argument -> Not supported.

	if (pState->ulProductCode==DJCONSOLE_PRODUCT_CODE)
	{
		// --- Record Input selection ---
		switch(eRecordInput)
		{
			case LINE_IN:
				ulValue = DJC_AUDIOCFG_ANALOG_IN_LINE;
				break;
			case MIC:
				ulValue = 0;
				break;
			case DIGITAL_COAXIAL:
				ulValue = DJC_AUDIOCFG_DIGITAL_IN|DJC_AUDIOCFG_DIGITAL_IN_COAXIAL;
				break;
			case DIGITAL_OPTICAL:
				ulValue = DJC_AUDIOCFG_DIGITAL_IN;
				break;
		}
		ulValue = (ulValue<<16)|
			(DJC_AUDIOCFG_DIGITAL_IN|DJC_AUDIOCFG_ANALOG_IN_LINE|DJC_AUDIOCFG_DIGITAL_IN_COAXIAL);

		if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_DJCONSOLE_CONFIG, &ulValue)<0)
			return false;
		pState->djConsole.eRecordingInput = eRecordInput;
	}
	else
	{
		if (eRecordInput==DIGITAL_COAXIAL || eRecordInput==DIGITAL_OPTICAL)
			return false;

		if (pState->ulProductCode==DJCONSOLE2_PRODUCT_CODE)
		{
			usValue = (eRecordInput==LINE_IN) ? DJMK2_AUDIOCFG_RECORD_INPUT_LINE_IN : 0;
			usValue = (usValue<<8) | DJMK2_AUDIOCFG_RECORD_INPUT_LINE_IN;

			if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_AUDIO_CONFIG, &usValue)<0)
				return false;
			pState->djConsoleMk2.bEnableInputMic = (eRecordInput==MIC);
		}
		else if (pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
		{
			// DJRMX_AUDIOCFG_RECORD_INPUT_LINE_IN
			usValue = (eRecordInput==LINE_IN) ? DJRMX_AUDIOCFG_RECORD_INPUT_LINE_IN : 0;
			usValue = (usValue<<8) | (DJRMX_AUDIOCFG_RECORD_INPUT_LINE_IN);

			if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_AUDIO_CONFIG, &usValue)<0)
				return false;
			pState->djConsoleRmx.bEnableInputMic = (eRecordInput==MIC);
		}
	}

	DEBUG_PRINT("bulk-SetRecordInput (eRecordInput=%d) with success", eRecordInput);
	return true;
}

// ---------------------------------------------------------------------------
// Function: SetTalkoverAttenuation
//
// Set Talkover attenuation in dB. For DJ Console Mk2 and Rmx set it to 0
// in order to disable talkover attenuation.
// ---------------------------------------------------------------------------
bool CBulkInterface::SetTalkoverAttenuation( int nTalkover, DJ_DEVICE_STATE *pState )		// DJC, Mk2, Rmx
{
	unsigned short usValue;

	if (!CheckIfSupported(pState, PROD_FLAGS_DJC|PROD_FLAGS_MK2|PROD_FLAGS_RMX))
		return false;	// Check Argument -> Not supported.

	usValue = nTalkover;	// DJC and Mk2, in dB.

	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_TALKOVER_ATT, &usValue)<0)
		return false;

	if (pState->ulProductCode==DJCONSOLE_PRODUCT_CODE)
		pState->djConsole.nTalkoverAttenuation = nTalkover;		
	else if (pState->ulProductCode==DJCONSOLE2_PRODUCT_CODE)
		pState->djConsoleMk2.nTalkoverAttenuation = nTalkover;
	else if (pState->ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
		pState->djConsoleRmx.nTalkoverAttenuation = nTalkover;

	DEBUG_PRINT("bulk-SetTalkoverAttenuation (nTalkover=%d) with success", nTalkover);
	return true;
}

// ---------------------------------------------------------------------------
// Function: SetOutputMainHwMix
//
// Set output main hardware mix usb1/2 + (none, mic or line in).
// ---------------------------------------------------------------------------
bool CBulkInterface::SetOutputMainHwMix( OUT_HWARE_MIX eHardwareMix, DJ_DEVICE_STATE *pState )	// Rmx
{
	unsigned short usValue;

	if (!CheckIfSupported(pState, PROD_FLAGS_RMX))
		return false;	// Check Argument -> Not supported.

	switch(eHardwareMix)
	{
		case NONE:
			usValue = 0;
			usValue = (usValue<<8) | DJRMX_AUDIOCFG_MAIN_OUT_MIX;
			break;
		case MIX_WITH_MIC:
			usValue = DJRMX_AUDIOCFG_MAIN_OUT_MIX;
			usValue = (usValue<<8) | DJRMX_AUDIOCFG_MAIN_OUT_MIX|DJRMX_AUDIOCFG_RECORD_INPUT_LINE_IN;
			break;
		case MIX_WITH_LINE_IN:
			usValue = DJRMX_AUDIOCFG_MAIN_OUT_MIX|DJRMX_AUDIOCFG_RECORD_INPUT_LINE_IN;
			usValue = (usValue<<8) | DJRMX_AUDIOCFG_MAIN_OUT_MIX|DJRMX_AUDIOCFG_RECORD_INPUT_LINE_IN;
			break;
	}

	// Set Audio Configuration.
	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_AUDIO_CONFIG, &usValue)<0)
		return false;

	// --- Set Output 1/2 Hardware Mix + Input record source if hardware mix is on ---
	pState->djConsoleRmx.eOutput12Mix = eHardwareMix;
	if (eHardwareMix==MIX_WITH_MIC)
		pState->djConsoleRmx.bEnableInputMic = true; 
	else if (eHardwareMix==MIX_WITH_LINE_IN)
		pState->djConsoleRmx.bEnableInputMic = false;

	DEBUG_PRINT("bulk-SetOutputMainHwMix (eHardwareMix=%d) with success", eHardwareMix);
	return true;
}

// ---------------------------------------------------------------------------
// Function: UseExternalMixer
//
// When enable, the internal mixer of the mk2 is disable.
// ---------------------------------------------------------------------------
bool CBulkInterface::UseExternalMixer( bool bUseExternalMixer, DJ_DEVICE_STATE *pState )	// Mk2
{
	unsigned short usValue;

	if (!CheckIfSupported(pState, PROD_FLAGS_MK2))
		return false;	// Check Argument -> Not supported.

	usValue = (bUseExternalMixer) ? DJMK2_AUDIOCFG_USE_EXTERNAL_MIXER : 0;
	usValue = (usValue<<8) | DJMK2_AUDIOCFG_USE_EXTERNAL_MIXER;

	// Set Audio Configuration.
	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_SET_AUDIO_CONFIG, &usValue)<0)
		return false;

	pState->djConsoleMk2.bUseExternalMixer = bUseExternalMixer;
	DEBUG_PRINT("bulk-UseExternalMixer (bUseExternalMixer=%d) with success", bUseExternalMixer);
	return true;
}

// ---------------------------------------------------------------------------
// Function: RegisterNetLink
//
// Register driver notification via NetLink.
// ---------------------------------------------------------------------------
bool CBulkInterface::RegisterNetLink( DJ_DEVICE_STATE *pState )
{
	unsigned long netlink_context;
	
	// register for notifications
	netlink_context = 0xdbdbffff; // We do not use netlink_context.

	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_REGISTER_FOR_NETLINK_DEVICE_NOTIFICATIONS, &netlink_context)<0)
	{
		DEBUG_PRINT("Bulk, Fail to register NET LINK.");
		return false;
	}

	DEBUG_PRINT("Custom Hercules Driver NET LINK registered.");
	return true;
}

// ---------------------------------------------------------------------------
// Function: UnRegisterNetLink
//
// Unregister driver notification via NetLink.
// ---------------------------------------------------------------------------
bool CBulkInterface::UnRegisterNetLink( DJ_DEVICE_STATE *pState )
{
	// register for notifications
	if (SingleBulkIoctl(pState->nBULK_FHandle, DJ_IOCTL_UNREGISTER_FOR_NETLINK_DEVICE_NOTIFICATIONS, NULL)<0)
	{
		DEBUG_PRINT("Bulk, Fail to unregister NET LINK.");
		return false;
	}

	DEBUG_PRINT("Custom Hercules Driver NET LINK unregistered.");
	return true;
}


