/////////////////////////////////////////////////////////////////////////////
// Name:        dmanager.cpp
// Purpose:     Device Manager provide a single set of functions for all
//		devices. Not all functions are available per DJ products.
//		Some functions are driven via different interfaces and are
//		re-routed in this class.
// Author:      Yann Hamiaux
// Modified by: -
// Created:     03/19/2008
// Copyright:   Written for Guillemot Corporation by Yann Hamiaux
// Licence:     The GNU General Public License.
/////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include "dmanager.h"

DJ_DEVICE_STATE		g_DEVICE_INFO[MAX_DEVICES];
volatile int		g_RescanDeviceList = 0;
wxDialog 		*g_pDlgParent;


CDManager::CDManager( void )
{
	int	i;

	// --- Empty Device Slots ---
	for(i=0;i<MAX_DEVICES;++i)
	{
		memset((void *) &g_DEVICE_INFO[i], 0, sizeof(DJ_DEVICE_STATE));
		g_DEVICE_INFO[i].nBULK_FHandle = -1;
	}
	g_RescanDeviceList = 1;
	m_nPreviousDeviceIndex = -1;

	m_pDriverNotification = NULL;
	m_pHidNotification = NULL;
}

CDManager::~CDManager( )
{
}

// ---------------------------------------------------------------------------
// Function: signal_handler
//
// SIGUSR1 are sent when device arrive or depart.
// ---------------------------------------------------------------------------
void CDManager::signal_handler(int sig, siginfo_t *sigInfo, void *uap)
{
	if (sigInfo==NULL)
		return;

	if (sigInfo->si_signo==SIGUSR1)
	{
		g_RescanDeviceList = 1;
	}
}

// ---------------------------------------------------------------------------
// Function: UpdateAlsa
//
// This function update the event callback on ALSA allowing a new
// event to be fire.
// ---------------------------------------------------------------------------
void CDManager::UpdateAlsa( DJ_DEVICE_STATE *pState )
{
	m_AlsaInterface.UpdateEventCallback(pState);
}

// ---------------------------------------------------------------------------
// Function: SelectDevice
//
// Close HID and custom driver notification on the previous device (if needed)
// then create new thread (as the previous threads has been destroy by
// deletion or stop) to handle the notification.
// ---------------------------------------------------------------------------
bool CDManager::SelectDevice( int nDeviceIndex )
{
	if (m_pHidNotification!=NULL)
	{
		DEBUG_PRINT("Deleting HID Notification thread.");
		m_pHidNotification->Delete();
		m_pHidNotification = NULL;
	}

	if (m_nPreviousDeviceIndex>=0)
	{
		DEBUG_PRINT("Unregister bulk interface NetLink notification.");
		m_BulkInterface.UnRegisterNetLink(&g_DEVICE_INFO[m_nPreviousDeviceIndex]);
		if (m_pDriverNotification!=NULL)
		{
			DEBUG_PRINT("Stopping Bulk driver notification thread.");
			m_pDriverNotification->Stop();
			m_pDriverNotification = NULL;
		}
	}


	if (nDeviceIndex>=0)
	{
		m_BulkInterface.RegisterNetLink(&g_DEVICE_INFO[nDeviceIndex]);
		g_DEVICE_INFO[nDeviceIndex].bBULK_Init = 
			m_BulkInterface.RetrieveState((DJ_DEVICE_STATE *) &g_DEVICE_INFO[nDeviceIndex]);

		DEBUG_PRINT("Creating Bulk driver notification thread.");
		m_pDriverNotification = new CDriverNotification();
		if (m_pDriverNotification!=NULL) {
			m_pDriverNotification->Create();
			m_pDriverNotification->Run(g_DEVICE_INFO[nDeviceIndex].nNetlinkUnit);
		}

		DEBUG_PRINT("Creating HID notification thread.");
		if (g_DEVICE_INFO[nDeviceIndex].ulProductCode==DJCONSOLE_PRODUCT_CODE || 
			g_DEVICE_INFO[nDeviceIndex].ulProductCode==DJCONSOLE2_PRODUCT_CODE)
		{
			m_pHidNotification = new CHIDNotification(&g_DEVICE_INFO[nDeviceIndex]);
			if (m_pHidNotification!=NULL) {
				m_pHidNotification->Create();
				m_pHidNotification->Run();
			}
		}
	}

	m_nPreviousDeviceIndex = nDeviceIndex;
	return true;
}

// ---------------------------------------------------------------------------
// Function: ScanDevices
//
// Reset device slots for each device not responding to IsDeviceAlive(), 
// set device slots for new device which has been detected.
// ---------------------------------------------------------------------------
bool CDManager::ScanDevices( void )
{
	bool bBulkAlloc[MAX_DEVICES+1];
	int i, iBulk;
	int nRetry;

	// --- Set allocation to false ---
	for(i=0;i<MAX_DEVICES+1;++i) {
		bBulkAlloc[i] = false;
	}

	// --- Check devices for Departure and identify bulk/midi allocation ---
	DEBUG_PRINT("Check if devices are still alive.");
	for(i=0;i<MAX_DEVICES;++i)
	{
		if (g_DEVICE_INFO[i].bPresent)
		{
			if (g_DEVICE_INFO[i].bBULK_Init)
			{
				if (!m_BulkInterface.IsDeviceAlive(&g_DEVICE_INFO[i]))
				{
					DEBUG_PRINT("Device %s has been remove.", g_DEVICE_INFO[i].szLocationID);
					m_AlsaInterface.Close(&g_DEVICE_INFO[i]);
					m_BulkInterface.Close(&g_DEVICE_INFO[i]);
					g_DEVICE_INFO[i].bPresent = false;
					memset(&g_DEVICE_INFO[i], 0, sizeof(DJ_DEVICE_STATE));
				}
			}

			if (g_DEVICE_INFO[i].bBULK_Init)
				bBulkAlloc[g_DEVICE_INFO[i].nBULK_FileIndex] = true;
		}
	}


	// --- Check devices for Arrival ---
	if (g_pDlgParent!=NULL)
		usleep(750000);
	iBulk = 0;
	i = 0;

	// --- Base on Bulk Interface open and allocate the device ---
	DEBUG_PRINT("Check device for arrival.");
	while(iBulk<MAX_DEVICES && i<MAX_DEVICES)
	{
		while(iBulk<MAX_DEVICES && bBulkAlloc[iBulk])
			++iBulk;

		while(i<MAX_DEVICES && g_DEVICE_INFO[i].bPresent)
			++i;

		m_BulkInterface.Open(iBulk, (DJ_DEVICE_STATE *) &g_DEVICE_INFO[i]);
		if (g_DEVICE_INFO[i].bBULK_Init)
		{
			g_DEVICE_INFO[i].bBULK_Init =
				m_BulkInterface.RetrieveState((DJ_DEVICE_STATE *) &g_DEVICE_INFO[i]);

			if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLLER_PRODUCT_CODE ||
				g_DEVICE_INFO[i].ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
			{
				// For these devices, no audio mixer exist.
				g_DEVICE_INFO[i].bPresent = true;
			}
			else
			{
				// --- ALSA may not be ready at this time ---
				// We have set a number of retry and delay by default.
				// Those number may need to be adjusted base on further tests.
				nRetry = 5;
				while(nRetry-->0)
				{
					if (m_AlsaInterface.Open((DJ_DEVICE_STATE *) &g_DEVICE_INFO[i]))
					{
						// Retrieve State: ALSA.
						g_DEVICE_INFO[i].bALSA_Init =
							m_AlsaInterface.RetrieveState((DJ_DEVICE_STATE *) &g_DEVICE_INFO[i]);
						if (g_DEVICE_INFO[i].bALSA_Init) {
							g_DEVICE_INFO[i].bPresent = true;
							nRetry = 0;

							DEBUG_PRINT("Device %s is or has been plug.", g_DEVICE_INFO[i].szLocationID);
						}
						else
							usleep(250000);
					}
					else
						usleep(500000);
				}
			}
		}
		++iBulk;
	}

	return true;
}


// ---------------------------------------------------------------------------
// Function: Initialize
// ---------------------------------------------------------------------------
bool CDManager::Initialize( wxDialog *pDlgParent )
{
	struct sigaction act;

	if (pDlgParent!=NULL)
	{
		g_pDlgParent = pDlgParent;
		sigemptyset(&act.sa_mask);
	        sigaddset(&act.sa_mask, SIGUSR1); // block USR1 when handler executes
		act.sa_flags = SA_SIGINFO | SA_RESTART;
		act.sa_sigaction = signal_handler;
		sigaction(SIGUSR1, &act, NULL);

//		sigemptyset(&act.sa_mask);
//	        sigaddset(&act.sa_mask, SIGPIPE); // block USR1 when USR2 handler executes
//		act.sa_flags = SA_SIGINFO | SA_RESTART;
//		act.sa_sigaction = signal_handler;
//		sigaction(SIGPIPE, &act, NULL);
	}

	return true;
}

// ---------------------------------------------------------------------------
// Function: Uninitialize
// ---------------------------------------------------------------------------
bool CDManager::Uninitialize( void )
{
	int i;

	DEBUG_PRINT("Uninitialize device manager.");
	for(i=0;i<MAX_DEVICES;++i)
	{
		if (g_DEVICE_INFO[i].bPresent)
		{
			DEBUG_PRINT("Clearing Device %s.", g_DEVICE_INFO[i].szLocationID);
			m_AlsaInterface.Close(&g_DEVICE_INFO[i]);
			m_BulkInterface.Close(&g_DEVICE_INFO[i]);
			g_DEVICE_INFO[i].bPresent = false;
			memset(&g_DEVICE_INFO[i], 0, sizeof(DJ_DEVICE_STATE));
		}
	}

	return true;
}

// ---------------------------------------------------------------------------
// Function: EnableDJMode
// ---------------------------------------------------------------------------
bool CDManager::EnableDJMode( bool bEnable, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.EnableDJMode(bEnable, pState);
}

// ---------------------------------------------------------------------------
// Function: SetMidiChannel
// ---------------------------------------------------------------------------
bool CDManager::SetMidiChannel( int nMidiChannel, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.SetMidiChannel(nMidiChannel, pState);
}

// ---------------------------------------------------------------------------
// Function: SetJogsSensitivity
// ---------------------------------------------------------------------------
bool CDManager::SetJogsSensitivity( JOGS_SENSITIVE eSensitivity, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.SetJogsSensitivity(eSensitivity, pState);
}

// ---------------------------------------------------------------------------
// Function: LockJogWheel
// ---------------------------------------------------------------------------
bool CDManager::LockJogWheel( JOG_WHEELS_DIS eJogWheelDisable, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.LockJogWheel(eJogWheelDisable, pState);
}

// ---------------------------------------------------------------------------
// Function: SetOutputMainHwMix
// ---------------------------------------------------------------------------
bool CDManager::SetOutputMainHwMix( OUT_HWARE_MIX eHardwareMix, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.SetOutputMainHwMix(eHardwareMix, pState);
}

// ---------------------------------------------------------------------------
// Function: SetRecordInput
// ---------------------------------------------------------------------------
bool CDManager::SetRecordInput( RECORD_INPUT eRecordInput, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.SetRecordInput(eRecordInput, pState);
}

// ---------------------------------------------------------------------------
// Function: SetMidiInput
// ---------------------------------------------------------------------------
bool CDManager::SetMidiInput( MIDI_INPUT_SEL eMidiInput, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.SetMidiInput(eMidiInput, pState);
}

// ---------------------------------------------------------------------------
// Function: SetHeadphoneToMainOut
// ---------------------------------------------------------------------------
bool CDManager::SetHeadphoneToMainOut( bool bHeadphoneToMain, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.SetHeadphoneToMainOut(bHeadphoneToMain,pState);
}

// ---------------------------------------------------------------------------
// Function: SetMute
// ---------------------------------------------------------------------------
bool CDManager::SetMute( DJ_DEVICE_STATE *pDeviceState, int nStereoChannel, bool bMute )
{
	if (nStereoChannel==1)
	{
		// For stereo channel 1, we need to set also the hardware mute via bulk interface.
		if (!m_BulkInterface.MuteChannel34(bMute, pDeviceState))
			return false;
	}

	return m_AlsaInterface.SetMute(pDeviceState, nStereoChannel, bMute);
}

// ---------------------------------------------------------------------------
// Function: SetVolume
// ---------------------------------------------------------------------------
bool CDManager::SetVolume( DJ_DEVICE_STATE *pDeviceState, int nChannel, unsigned short usVolume )
{
	return m_AlsaInterface.SetVolume(pDeviceState, nChannel, usVolume);
}

// ---------------------------------------------------------------------------
// Function: EnableTalkover
// ---------------------------------------------------------------------------
bool CDManager::EnableTalkover( bool bEnable, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.EnableTalkover(bEnable,pState);
}

// ---------------------------------------------------------------------------
// Function: SetTalkoverAttenuation
// ---------------------------------------------------------------------------
bool CDManager::SetTalkoverAttenuation( int nTalkover, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.SetTalkoverAttenuation(nTalkover,pState);
}

// ---------------------------------------------------------------------------
// Function: EnableDJMouse
// ---------------------------------------------------------------------------
bool CDManager::EnableDJMouse( bool bEnable, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.EnableDJMouse(bEnable, pState);
}

// ---------------------------------------------------------------------------
// Function: UseExternalMixer
// ---------------------------------------------------------------------------
bool CDManager::UseExternalMixer( bool bUseExternalMixer, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.UseExternalMixer(bUseExternalMixer, pState);
}

// ---------------------------------------------------------------------------
// Function: SetCrossFaderCurve
// ---------------------------------------------------------------------------
bool CDManager::SetCrossFaderCurve( CROSSFADER_SEL eCurve, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.SetCrossFaderCurve( eCurve, pState );
}

// ---------------------------------------------------------------------------
// Function: SetSampleRate
// ---------------------------------------------------------------------------
bool CDManager::SetSampleRate( int nSampleRate, DJ_DEVICE_STATE *pState )
{
	return m_BulkInterface.SetSampleRate( nSampleRate, pState );
}

// ---------------------------------------------------------------------------
// Function: SetLockState
// ---------------------------------------------------------------------------
bool CDManager::SetLockState( int nDeck, bool bLock, DJ_DEVICE_STATE *pState )
{
	if (pState->ulProductCode!=DJCONSOLE2_PRODUCT_CODE)
		return false;

	pState->djConsoleMk2.bLockAudioSource[nDeck] = bLock;
	m_BulkInterface.UpdateHIDSet_DJConsoleMk2(pState);
	return true;
}


