/////////////////////////////////////////////////////////////////////////////
// Name:        drivernotification.cpp
// Purpose:     wxThread derived class to handle Hercules custom driver
//		notification (called BULK driver). Notification are
//		provided via the Net Link interface.
// Author:      Yann Hamiaux
// Modified by: -
// Created:     15/09/2008
// Copyright:   Written for Guillemot Corporation by Yann Hamiaux
// Licence:     The GNU General Public License v2+.
/////////////////////////////////////////////////////////////////////////////
#include "drivernotification.h"
#include "commondefs.h"
#include <sys/socket.h>
#include <linux/netlink.h>
#include "callback.h"

extern DJ_DEVICE_STATE	g_DEVICE_INFO[MAX_DEVICES];
#define MAX_PAYLOAD 1024  // maximum payload size.


CDriverNotification::CDriverNotification( void )
{
	m_nNetLinkUnit = NETLINK_UNIT_INVALID_VALUE;
}

CDriverNotification::~CDriverNotification()
{
}

// ---------------------------------------------------------------------------
// Function: Run
//
// Set the Net Link unit receive by the driver to register properly and call
// the normal Run function in wxThread.
// ---------------------------------------------------------------------------
bool CDriverNotification::Run( int nNetLinkUnit )
{
	m_nNetLinkUnit = nNetLinkUnit;

	if (pipe(&m_pipe_notification[0])!=0)
		return false;

	wxThread::Run();
	return true;
}

// ---------------------------------------------------------------------------
// Function: Stop
//
// Write to the pipe that the thread need to bail out.
// ---------------------------------------------------------------------------
bool CDriverNotification::Stop( void )
{
	char buf;
	int rc;

	buf = MSG_THREAD_BAIL;
	rc  = write(m_pipe_notification[1],&buf,sizeof(buf));
	return true;
}

// ---------------------------------------------------------------------------
// Function: Entry
//
// Main loop where we receive and process driver notifications.
// ---------------------------------------------------------------------------
void* CDriverNotification::Entry()
{
	struct sockaddr_nl 	 src_addr, dest_addr;
	struct nlmsghdr 	*nlh = NULL;
	struct iovec 		 iov;
	struct msghdr 		 msg;
	fd_set			 rd_set, ex_set;
	unsigned char 		*current_msg = NULL;
	unsigned char 		 sig_buf;
	unsigned long 		 current_msg_len=0;
	int 			 rc, max_fd;
	int 			 sock_fd=-1;
	struct netlink_msg_header* netlink_msg;
	struct control_change_data* ctrlchange_msg;
	int			 netlink_msg_size;
	int			 i, nValue;
	int 			 nLock, nCurveMul, nCurveOffset;

	if (m_nNetLinkUnit==NETLINK_UNIT_INVALID_VALUE)
	{
		// If net link unit is invalid, the driver could not allocate a net link unit.
		// In that case, the socket could not be created to get notification.
		
		// We could poll the device state, in case if there is no NETLINK unit
		// available. For now, we assume the driver could always found one available.
		return Wait();
	}

	sock_fd = socket(PF_NETLINK, SOCK_DGRAM, m_nNetLinkUnit);
	if (sock_fd == -1)
		goto kernel_notification_thread_exit;

	memset(&src_addr, 0, sizeof(src_addr));
	src_addr.nl_family = AF_NETLINK;
	src_addr.nl_pid = getpid();	// self pid
	src_addr.nl_groups = 0;		// not in mcast groups

	if (bind(sock_fd, (struct sockaddr*)&src_addr, sizeof(src_addr))==-1)
		goto kernel_notification_thread_exit;

	memset(&dest_addr, 0, sizeof(dest_addr));
	dest_addr.nl_family = AF_NETLINK;
	dest_addr.nl_pid = 0;		// For Linux Kernel
	dest_addr.nl_groups = 0;	// unicast

	nlh=(struct nlmsghdr *)calloc(1,NLMSG_SPACE(MAX_PAYLOAD));
	if (nlh==NULL)
		goto kernel_notification_thread_exit;

	// we copy received data here, until the packet is complet.
	current_msg = (unsigned char*)calloc(1,MAX_PAYLOAD);
	if (current_msg==NULL)
		goto kernel_notification_thread_exit;

	iov.iov_base = (void *)nlh;
	iov.iov_len = NLMSG_SPACE(MAX_PAYLOAD); // nlh->nlmsg_len;
	msg.msg_name = (void *)&dest_addr;
	msg.msg_namelen = sizeof(dest_addr);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;


	while(1)
	{
process_next_notification:
		// prepare ourselves to call select.
		// do for each call to select, zero our descriptor sets.
		FD_ZERO(&rd_set);
		FD_ZERO(&ex_set);

		FD_SET(sock_fd, &rd_set);
		FD_SET(m_pipe_notification[0], &rd_set);
		FD_SET(sock_fd, &ex_set);
		FD_SET(m_pipe_notification[0], &ex_set);

		// extract the maximum descriptor for select
		if (sock_fd > m_pipe_notification[0])
			max_fd = sock_fd;
		else
			max_fd = m_pipe_notification[0];

		// block until we receive data from the driver, or the pipe
		rc = select(max_fd+1, &rd_set, NULL, &ex_set, NULL);
		if (rc<0)
			goto process_next_notification;

		// check first if we are asked to bail.
		if (FD_ISSET(m_pipe_notification[0], &rd_set))
		{
			rc=read(m_pipe_notification[0],&sig_buf,sizeof(sig_buf));
			if (rc==1 && sig_buf==MSG_THREAD_BAIL)
				goto kernel_notification_thread_exit;
		}
		else if (!FD_ISSET(sock_fd, &rd_set))
			goto process_next_notification;

		// check for messages from the device driver.
		rc = recvmsg(sock_fd, &msg, 0);
		if (rc <= NLMSG_LENGTH(0))
			goto process_next_notification;

		memcpy(current_msg+current_msg_len,NLMSG_DATA(nlh),NLMSG_PAYLOAD(nlh, 0));
		current_msg_len += NLMSG_PAYLOAD(nlh, 0);
		if (current_msg_len < sizeof(struct netlink_msg_header) )
			goto process_next_notification;

		// we have the header, check if we have the message too.
		netlink_msg = (struct netlink_msg_header*)current_msg;
		netlink_msg_size = netlink_msg->bytes_to_follow+netlink_msg->header_size;

		if (current_msg_len < (unsigned long) netlink_msg_size)
			goto process_next_notification;

		// process control change notification.
		if (netlink_msg->msg_magic_number != MAGIC_NUM || netlink_msg->msg_id!=MSG_CONTROL_CHANGE)
			goto skip_notification;

		ctrlchange_msg = (struct control_change_data*) &current_msg[netlink_msg->header_size];
		if (netlink_msg->bytes_to_follow<sizeof(struct control_change_data))
			goto skip_notification;

		// ---  ---
		i = 0;
		while(i<MAX_DEVICES && strcmp(g_DEVICE_INFO[i].szLocationID,
			(char *) ctrlchange_msg->location_id)!=0) {
			++i;
		}

		if (i>=MAX_DEVICES)
			goto skip_notification;

		switch(ctrlchange_msg->control_id)
		{
		case CTRL_CHG_JOG_WHEEL_LOCK:
			if(ctrlchange_msg->control_value==0x0000) {
				if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
					g_DEVICE_INFO[i].djConsoleRmx.eJogWheelsLock = ENABLE;
				else if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
					g_DEVICE_INFO[i].djControlSteel.eJogWheelsLock = ENABLE;
			}
			else if(ctrlchange_msg->control_value==0x0001) {
				if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
					g_DEVICE_INFO[i].djConsoleRmx.eJogWheelsLock = DISABLE_A;
				else if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
					g_DEVICE_INFO[i].djControlSteel.eJogWheelsLock = DISABLE_A;
			}
			else if(ctrlchange_msg->control_value==0x0100) {
				if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
					g_DEVICE_INFO[i].djConsoleRmx.eJogWheelsLock = DISABLE_B;
				else if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
					g_DEVICE_INFO[i].djControlSteel.eJogWheelsLock = DISABLE_B;
			}
			else if(ctrlchange_msg->control_value==0x0101) {
				if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
					g_DEVICE_INFO[i].djConsoleRmx.eJogWheelsLock = DISABLE_BOTH;
				else if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
					g_DEVICE_INFO[i].djControlSteel.eJogWheelsLock = DISABLE_BOTH;
			}
			break;
		case CTRL_CHG_JOG_WHEEL_SENS:
			if(ctrlchange_msg->control_value==0x0000) {
				if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
					g_DEVICE_INFO[i].djConsoleRmx.eJogWheelsSensitivity = NORMAL;
				else if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
					g_DEVICE_INFO[i].djControlSteel.eJogWheelsSensitivity = NORMAL;
			}
			else if(ctrlchange_msg->control_value==0x0001) {
				if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
					g_DEVICE_INFO[i].djConsoleRmx.eJogWheelsSensitivity = LESS;
				else if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
					g_DEVICE_INFO[i].djControlSteel.eJogWheelsSensitivity = LESS;
			}
			else if(ctrlchange_msg->control_value==0x0002) {
				if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
					g_DEVICE_INFO[i].djConsoleRmx.eJogWheelsSensitivity = LEAST;
				else if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
					g_DEVICE_INFO[i].djControlSteel.eJogWheelsSensitivity = LEAST;
			}
			break;
		case CTRL_CHG_MIDI_CHANNEL:
			if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLLER_PRODUCT_CODE)
				g_DEVICE_INFO[i].djControlMP3.nMidiChannel = ctrlchange_msg->control_value;
			else if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
				g_DEVICE_INFO[i].djConsoleRmx.nMidiChannel = ctrlchange_msg->control_value;
			else if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLSTEEL_PRODUCT_CODE)
				g_DEVICE_INFO[i].djControlSteel.nMidiChannel = ctrlchange_msg->control_value;
			break;
		case CTRL_CHG_TALKOVER_ATTEN:
			nValue = ctrlchange_msg->control_value;
			if (nValue<=4)
				nValue = 3;
			else if (nValue<=7)
				nValue = 6;
			else
				nValue = 9;

			if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLE_PRODUCT_CODE)
				g_DEVICE_INFO[i].djConsole.nTalkoverAttenuation = nValue;
			else if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLE2_PRODUCT_CODE)
				g_DEVICE_INFO[i].djConsoleMk2.nTalkoverAttenuation = nValue;
			else if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
				g_DEVICE_INFO[i].djConsoleRmx.nTalkoverAttenuation = nValue;
			break;
		case CTRL_CHG_TALKOVER_ENABLE:
			if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLE_PRODUCT_CODE)
				g_DEVICE_INFO[i].djConsole.bTalkoverAttenuation = (ctrlchange_msg->control_value!=0);
			else if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLE2_PRODUCT_CODE)
				g_DEVICE_INFO[i].djConsoleMk2.bTalkoverAttenuation = (ctrlchange_msg->control_value!=0);
			else if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
				g_DEVICE_INFO[i].djConsoleRmx.bTalkoverAttenuation = (ctrlchange_msg->control_value!=0);
			break;
		case CTRL_CHG_DJ1_DEVICE_CONFIG:
			g_DEVICE_INFO[i].djConsole.bEnableDJMode = 
				(ctrlchange_msg->control_value&DJC_AUDIOCFG_ADVANCED_MODE);
			g_DEVICE_INFO[i].djConsole.bHeadphoneOnMaster = 
				(ctrlchange_msg->control_value&DJC_AUDIOCFG_HEADPHONE_INPUT_SEL);
			g_DEVICE_INFO[i].djConsole.eMidiInput = (ctrlchange_msg->control_value & 
				DJC_AUDIOCFG_MIDI_IN_BUTTONS) ? DJ_CONSOLE_BUTTONS : MIDI_IN_PLUG;
			g_DEVICE_INFO[i].djConsole.bMute[1] = (ctrlchange_msg->control_value & 
				DJC_AUDIOCFG_MUTE_REAR_3_4);

			if (ctrlchange_msg->control_value&DJC_AUDIOCFG_DIGITAL_IN) {
				g_DEVICE_INFO[i].djConsole.eRecordingInput = (ctrlchange_msg->control_value & 
					DJC_AUDIOCFG_DIGITAL_IN_COAXIAL) ? DIGITAL_COAXIAL : DIGITAL_OPTICAL;
			}
			else {
				g_DEVICE_INFO[i].djConsole.eRecordingInput = (ctrlchange_msg->control_value & 
					DJC_AUDIOCFG_ANALOG_IN_LINE) ? LINE_IN : MIC;
			}
			break;
		case CTRL_CHG_AUDIO_CONFIG:
			if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLERMX_PRODUCT_CODE)
			{
				g_DEVICE_INFO[i].djConsoleRmx.bMute[1] = 
					(ctrlchange_msg->control_value&DJRMX_AUDIOCFG_MUTE_34);
				g_DEVICE_INFO[i].djConsoleRmx.bEnableInputMic = 
					!(ctrlchange_msg->control_value&DJRMX_AUDIOCFG_RECORD_INPUT_LINE_IN);
				if (ctrlchange_msg->control_value&DJRMX_AUDIOCFG_MAIN_OUT_MIX)
				{
					g_DEVICE_INFO[i].djConsoleRmx.eOutput12Mix = 
						(g_DEVICE_INFO[i].djConsoleRmx.bEnableInputMic) ? 
						MIX_WITH_MIC : MIX_WITH_LINE_IN;
				}
				else
					g_DEVICE_INFO[i].djConsoleRmx.eOutput12Mix = NONE;
			}
			else if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLE2_PRODUCT_CODE)
			{
				g_DEVICE_INFO[i].djConsoleMk2.bMute[1] = 
					(ctrlchange_msg->control_value&DJMK2_AUDIOCFG_HW_MUTE_CH_3_4);
				g_DEVICE_INFO[i].djConsoleMk2.bEnableInputMic = 
					!(ctrlchange_msg->control_value&DJMK2_AUDIOCFG_RECORD_INPUT_LINE_IN);
				g_DEVICE_INFO[i].djConsoleMk2.bUseExternalMixer = 
					(ctrlchange_msg->control_value&DJMK2_AUDIOCFG_USE_EXTERNAL_MIXER);
				g_DEVICE_INFO[i].djConsoleMk2.eAudioSourceSelection[0] = 
					(ctrlchange_msg->control_value&DJMK2_AUDIOCFG_DECK_A_OUTPUT_1_2)
					? USB_PC_OUT : MK2_AUDIO_INPUT;
				g_DEVICE_INFO[i].djConsoleMk2.eAudioSourceSelection[1] = 
					(ctrlchange_msg->control_value&DJMK2_AUDIOCFG_DECK_B_OUTPUT_3_4)
					? USB_PC_OUT : MK2_AUDIO_INPUT;
			}
			break;
		case CTRL_CHG_DJ_MOUSE_ENABLE:
			if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLE_PRODUCT_CODE)
				g_DEVICE_INFO[i].djConsole.bEnableDJMouse = (ctrlchange_msg->control_value!=0);
			else if (g_DEVICE_INFO[i].ulProductCode==DJCONTROLLER_PRODUCT_CODE)
				g_DEVICE_INFO[i].djControlMP3.bEnableDJMouse = (ctrlchange_msg->control_value!=0);
			else if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLE2_PRODUCT_CODE)
				g_DEVICE_INFO[i].djConsoleMk2.bEnableDJMouse = (ctrlchange_msg->control_value!=0);
			break;
		case CTRL_CHG_SAMPLE_RATE:
			if (g_DEVICE_INFO[i].ulProductCode==DJCONSOLE_PRODUCT_CODE)
				g_DEVICE_INFO[i].djConsole.nSampleRate = ctrlchange_msg->control_value;
			break;
		case CTRL_CHG_XFADER_LOCK:
		case CTRL_CHG_XFADER_STYLE:
			// --- Retrieve Lock, Multiplier and Offset for cross-fader setting ---
			if (ctrlchange_msg->control_id==CTRL_CHG_XFADER_LOCK)
			{
				nLock = (ctrlchange_msg->control_value>>0)&0xFFFF;
				nCurveMul = (ctrlchange_msg->control_value>>16)&0xFF;
				nCurveOffset = (ctrlchange_msg->control_value>>24)&0xFF;
			}
			else
			{
				nLock = (ctrlchange_msg->control_value>>16)&0xFFFF;
				nCurveMul = (ctrlchange_msg->control_value>>0)&0xFF;
				nCurveOffset = (ctrlchange_msg->control_value>>8)&0xFF;
			}

			// --- Determine one of the 6 presets which is available ---
			if (nLock==0)
			{
				if (nCurveMul<13)
					g_DEVICE_INFO[i].djConsoleMk2.eCrossFaderCurve = DIPPED;
				else if (nCurveMul<20)
					g_DEVICE_INFO[i].djConsoleMk2.eCrossFaderCurve = BEATMIX;
				else
					g_DEVICE_INFO[i].djConsoleMk2.eCrossFaderCurve = SCRATCH;
			}
			else if (nLock==0x10)
				g_DEVICE_INFO[i].djConsoleMk2.eCrossFaderCurve = LOCK_ON_A;
			else if (nLock==0x20)
				g_DEVICE_INFO[i].djConsoleMk2.eCrossFaderCurve = LOCK_ON_B;
			else
				g_DEVICE_INFO[i].djConsoleMk2.eCrossFaderCurve = LOCK_ON_CENTER;
			break;
		case CTRL_CHG_SHIFT_MODE_STATE:
			break;
		case CTRL_CHG_FX_STATE:
			break;
		case CTRL_CHG_OUTPUT_CONTROL:
			break;
		}
		g_DEVICE_INFO[i].bStateChange = true;

skip_notification:
		// remove used bytes reserved for our netlink_msg
		if (current_msg_len == (unsigned long) netlink_msg_size)
			current_msg_len = 0;
		else
		{
			current_msg_len -= netlink_msg_size;
			for(i=0;i<(int) (MAX_PAYLOAD - current_msg_len);++i)
				current_msg[i] = current_msg[i+current_msg_len];
		}
	}

kernel_notification_thread_exit:
	if (nlh!=NULL)
		free(nlh);

	if (current_msg!=NULL)
		free(current_msg);

	if (sock_fd!=-1)
		close(sock_fd);

	return Wait();
}


