/////////////////////////////////////////////////////////////////////////////
// Name:        hdjcpl.cpp
// Purpose:     Application to control the hardware of Hercules DJ products.
// Author:      Yann Hamiaux
// Modified by: -
// Created:     02/01/2008
// Copyright:   Written for Guillemot Corporation by Yann Hamiaux
// Licence:     The GNU General Public License.
/////////////////////////////////////////////////////////////////////////////

#include "wx/wx.h" 
#include "wx/image.h"
#include "wx/tipwin.h"
#include <sys/utsname.h>
#include <sys/stat.h>
#include <string.h>

#include "binreloc.h"
#include "devnulldlg.h"
#include "devdjcdlg.h"
#include "devmk2dlg.h"
#include "devmp3dlg.h"
#include "devrmxdlg.h"
#include "devsteeldlg.h"
#include "dmanager.h"
#include "localization.h"
#include "hdjcpl_icon.h"


#define SAFE_DELETE(p); { if (p) {delete p; p = NULL;} }

#define TIMER_ID 			1
#define TIMER_CHECK_DEV_CHANGE_ID	2

#define TIMER_CHECK_DEV_CHANGE_DELAY	30

#define IDM_DEVICE_INDEX_0		0
#define IDM_DEVICE_INDEX_1		1
#define IDM_DEVICE_INDEX_2		2
#define IDM_DEVICE_INDEX_3		3
#define IDM_DEVICE_INDEX_4		4
#define IDM_DEVICE_INDEX_5		5
#define IDM_DEVICE_INDEX_6		6
#define IDM_DEVICE_INDEX_7		7
#define IDM_DEVICE_INDEX_8		8
#define IDM_DEVICE_INDEX_9		9
#define IDM_DEVICE_INDEX_10		10
#define IDM_DEVICE_INDEX_11		11
#define IDM_DEVICE_INDEX_12		12
#define IDM_DEVICE_INDEX_13		13
#define IDM_DEVICE_INDEX_14		14
#define IDM_DEVICE_INDEX_15		15

char *GetDeviceName( int nDevType );
char *g_pDeviceSelect = NULL;
bool  g_bDebug = false;


CLocalization 	g_LOCALIZATION;

class MyDlg;
class MyApp: public wxApp
{
public:
    virtual bool OnInit();
    const wxImage& GetImage() const { return m_image; }
    const wxImage& GetImageClose() const { return m_imageBtn; }

private:
    MyDlg *m_frame;
    wxImage m_image, m_imageBtn;
};


class MyDlg: public wxDialog
{
public:
    	MyDlg(const wxString& title, const wxPoint& pos, const wxSize& size);
	~MyDlg();
    	bool Destroy( void );
    	void OnPaint (wxPaintEvent & event);
    	void OnMouseMotion (wxMouseEvent & event);
	void OnMouseLeftUp(wxMouseEvent & event);
	void OnMouseRightUp(wxMouseEvent & event);
	void OnMouseLeftDown(wxMouseEvent & event);
	void OnMouseDClick(wxMouseEvent &event);
	void OnTimer(wxTimerEvent& event);
	void OnCheckDevChange(wxTimerEvent& event);
	void OnErase( wxEraseEvent & event);
	void OnSizing( wxSizeEvent &event );
	void OnIDM_DeviceIndex0(wxCommandEvent &event) {ChangeDevice(0);}
	void OnIDM_DeviceIndex1(wxCommandEvent &event) {ChangeDevice(1);}
	void OnIDM_DeviceIndex2(wxCommandEvent &event) {ChangeDevice(2);}
	void OnIDM_DeviceIndex3(wxCommandEvent &event) {ChangeDevice(3);}
	void OnIDM_DeviceIndex4(wxCommandEvent &event) {ChangeDevice(4);}
	void OnIDM_DeviceIndex5(wxCommandEvent &event) {ChangeDevice(5);}
	void OnIDM_DeviceIndex6(wxCommandEvent &event) {ChangeDevice(6);}
	void OnIDM_DeviceIndex7(wxCommandEvent &event) {ChangeDevice(7);}
	void OnIDM_DeviceIndex8(wxCommandEvent &event) {ChangeDevice(8);}
	void OnIDM_DeviceIndex9(wxCommandEvent &event) {ChangeDevice(9);}
	void OnIDM_DeviceIndex10(wxCommandEvent &event) {ChangeDevice(10);}
	void OnIDM_DeviceIndex11(wxCommandEvent &event) {ChangeDevice(11);}
	void OnIDM_DeviceIndex12(wxCommandEvent &event) {ChangeDevice(12);}
	void OnIDM_DeviceIndex13(wxCommandEvent &event) {ChangeDevice(13);}
	void OnIDM_DeviceIndex14(wxCommandEvent &event) {ChangeDevice(14);}
	void OnIDM_DeviceIndex15(wxCommandEvent &event) {ChangeDevice(15);}
    DECLARE_EVENT_TABLE()

protected:
private:
	void 	ChangeDevice( int nIndex );
	int	m_x, m_y, m_nDeviceIndex;
	bool 	m_bCloseDown, m_bCloseOver;

	// --- Only current resource ---
	wxImage		 m_dashboard, m_dashboardAlpha;
	wxBitmap	*m_pBmpDashboard, *m_pBmpDashboardAlpha;
	wxMemoryDC	 m_memDCDashboard, m_memDCDashboardAlpha;

	CGDialog	*m_pDialog;
	CDManager	 m_DManager;
	wxTimer		 m_timer, m_timerDevChange;
	char 		 m_szBasePath[512];
};

#define PRG_ARG_HELP           "--help"          // Display help page.
#define PRG_ARG_HELP2          "-?"              // Same as '--help'
#define PRG_ARG_DEBUG          "-debug"          // Output debug strings to console (which may be re-direct to a file)
#define PRG_ARG_DEVICE         "-device"         // Select a specific device (-device=DEVICE_PHYSICAL_LOCATION)
#define PRG_ARG_LANGUAGE       "-lang"           // Specify the user-language (-lang=LANGUAGE_CODE)
#define PRG_ARG_LICENSE        "-license"        // Display license and version number (same as -version)
#define PRG_ARG_DEVICE_LIST    "-ls"             // List detected devices with their physical location
#define PRG_ARG_MULTI_INSTANCE "-multi-instance" // Allow multiple instances of the application
#define PRG_ARG_VERSION        "-version"        // Display license and version number (same as -license)

// ---------------------------------------------------------------------------
// Function: printUsageArg
//
// This function print the argument prefix (first and second if specify)
// with its comment space evenly.
// ---------------------------------------------------------------------------
void printUsageArg( const char *prefix, const char *prefix2, const char *comment )
{
	char string[260];
	int i;
	
	if (prefix2!=NULL)
		sprintf(string, "  %s, %s", prefix, prefix2);
	else
		sprintf(string, "  %s", prefix);

	i = 20-strlen(string);
	while(i-->0)
		strcat(string, " ");
	strcat(string, comment);
	printf("%s\n", string);
}

// ---------------------------------------------------------------------------
// Function: processCount
//
// Return the number of process launched which match the path of this
// process (including ourself).
// ---------------------------------------------------------------------------
int processCount( void )
{
	char	szOurProcIdExe[260], szProcExePath[260];
	char	szOurExecPath[512], szExecPath[512];
	DIR	*pDirHandle;
	struct dirent 	*pDirEntry;
	int 	nCount = 0;

	sprintf(szOurProcIdExe, "/proc/%d/exe", getpid());
	memset(szOurExecPath, 0, 512);
	readlink(szOurProcIdExe, szOurExecPath, 512);
	pDirHandle = opendir("/proc");
	if (pDirHandle!=NULL) {
		do 
		{
			pDirEntry = readdir(pDirHandle);
			if (pDirEntry!=NULL) 
			{
				sprintf(szProcExePath, "/proc/%s/exe", pDirEntry->d_name);
				// Try to read the link.
				memset(szExecPath, 0, 512);
				if (readlink(szProcExePath, szExecPath, 512)>0) 
				{
					if (strcmp(szExecPath, szOurExecPath)==0 && 
						strcmp("/proc/self/exe", szProcExePath)!=0) 
					{
						++nCount;
					}
				}
			}
		} while (pDirEntry!=NULL);
		closedir(pDirHandle);
	}
	DEBUG_PRINT("We found %d similar process running (including ourself).", nCount);
	return nCount;
}

// ---------------------------------------------------------------------------
// Function: OnInit
//
// Main function in which we first look for program arguments and may
// start the application if condition are met.
// ---------------------------------------------------------------------------
bool MyApp::OnInit()
{
	bool bDisplayHelp = false;
	bool bDisplayDeviceList = false;
	bool bDisplayLicense = false;
	bool bAllowMultipleInstance = false;
	char *pUnrecognizeArg = NULL;
	char *pLangCode = NULL;
	char *pDevice   = NULL;
	char strTemp[260], strTemp2[260];
	int i, j;

	// --- Parse Arguments ---
	for(i=1;i<argc;++i)
	{
		if (strcasecmp(argv[i], PRG_ARG_HELP)==0 || strcasecmp(argv[i], PRG_ARG_HELP2)==0)
			bDisplayHelp = true;
		else if (strcasecmp(argv[i], PRG_ARG_DEBUG)==0)
			g_bDebug = true;
		else if (strcasecmp(argv[i], PRG_ARG_DEVICE_LIST)==0)
			bDisplayDeviceList = true;
		else if (strcasecmp(argv[i], PRG_ARG_LICENSE)==0 || strcasecmp(argv[i], PRG_ARG_VERSION)==0)
			bDisplayLicense = true;
		else if (strcasecmp(argv[i], PRG_ARG_MULTI_INSTANCE)==0)
			bAllowMultipleInstance = true;
		else if (strlen(PRG_ARG_LANGUAGE)<strlen(argv[i]) &&
			strncasecmp(argv[i], PRG_ARG_LANGUAGE, strlen(PRG_ARG_LANGUAGE))==0)
			pLangCode = &argv[i][strlen(PRG_ARG_LANGUAGE)+1];
		else if (strlen(PRG_ARG_DEVICE)<strlen(argv[i]) &&
			strncasecmp(argv[i], PRG_ARG_DEVICE, strlen(PRG_ARG_DEVICE))==0)
			pDevice = &argv[i][strlen(PRG_ARG_DEVICE)+1];
		else if (pUnrecognizeArg==NULL)
			pUnrecognizeArg = argv[i];
	}

	DEBUG_PRINT("Parsing arguments... completed.");

	// --- Set Application Language ---
	g_LOCALIZATION.SetLanguageID(getenv("LANG"));
	if (pLangCode!=NULL)
	{
		if (!g_LOCALIZATION.SetLanguageID(pLangCode))
		{
			sprintf(strTemp, "%s\n", g_LOCALIZATION.GetString(IDS_WARNING_LANG_NOT_SUPPORTED));
			printf(strTemp, pLangCode);
		}
	}

	// --- Display Help ---
	if (bDisplayHelp)
	{
		printf("%s %s [%s]\n", g_LOCALIZATION.GetString(IDS_USAGE), PROGRAM_NAME,
			g_LOCALIZATION.GetString(IDS_OPTION));
		printf("%s, %s %s\n", g_LOCALIZATION.GetString(IDS_HERCULES_CONTROL_PANEL), 
			g_LOCALIZATION.GetString(IDS_VERSION), INFO_CONTROL_PANEL);
		printf("\n");

		// --- print possible argument list ---
		printUsageArg(PRG_ARG_HELP2, PRG_ARG_HELP, g_LOCALIZATION.GetString(IDS_GIVE_THIS_HELP));
		printUsageArg(PRG_ARG_DEBUG, NULL, g_LOCALIZATION.GetString(IDS_OUTPUT_DEBUG_STRINGS));
		sprintf(strTemp, "%s=PHYSLOC", PRG_ARG_DEVICE);
		printUsageArg(strTemp, NULL, g_LOCALIZATION.GetString(IDS_SET_CURRENT_DEVICE));
		sprintf(strTemp, "%s=LANG", PRG_ARG_LANGUAGE);
		printUsageArg(strTemp, NULL, g_LOCALIZATION.GetString(IDS_SPECIFY_USER_LANGUAGE));
		sprintf(strTemp, "%s (%s)", g_LOCALIZATION.GetString(IDS_DISPLAY_LICENSE),
			g_LOCALIZATION.GetString(IDS_SAME_AS));
		sprintf(strTemp2, strTemp, PRG_ARG_VERSION);
		printUsageArg(PRG_ARG_LICENSE, NULL, strTemp2);
		printUsageArg(PRG_ARG_DEVICE_LIST, NULL, g_LOCALIZATION.GetString(IDS_LIST_DETECTED_DEVICES));
		printUsageArg(PRG_ARG_MULTI_INSTANCE, NULL, g_LOCALIZATION.GetString(IDS_ALLOW_MULTIPLE_INSTANCES));
		sprintf(strTemp, "%s (%s)", g_LOCALIZATION.GetString(IDS_DISPLAY_LICENSE),
			g_LOCALIZATION.GetString(IDS_SAME_AS));
		sprintf(strTemp2, strTemp, PRG_ARG_LICENSE);
		printUsageArg(PRG_ARG_VERSION, NULL, strTemp2);

		printf("\n%s\n", g_LOCALIZATION.GetString(IDS_REPORT_BUGS));
		exit(0);
	}

	// --- Display License ---
	if (bDisplayLicense)
	{
		printf("%s %s\n", PROGRAM_NAME, INFO_CONTROL_PANEL);
		printf("Copyright %s\n", g_LOCALIZATION.GetString(IDS_COPYRIGHT));
		printf("\n%s\n\n", g_LOCALIZATION.GetString(IDS_LICENSE_CONTACT));
		printf("%s %s \n\n", 
			g_LOCALIZATION.GetString(IDS_THIS_IS_FREE_SOFTWARE),
			g_LOCALIZATION.GetString(IDS_LICENSE_GNU_NOTE_1));
		printf("%s\n<http://www.gnu.org/licenses/gpl.html>\n\n", g_LOCALIZATION.GetString(IDS_LICENSE_GNU_NOTE_2));
		printf("%s\n", g_LOCALIZATION.GetString(IDS_WRITTEN_BY));
		exit(0);
	}

	// --- Unrecognize argument ---
	if (pUnrecognizeArg!=NULL)
	{
		sprintf(strTemp, "%s: %s.\n", PROGRAM_NAME,
			g_LOCALIZATION.GetString(IDS_INVALID_OPTION));
		printf(strTemp, pUnrecognizeArg);
		sprintf(strTemp, "%s\n", g_LOCALIZATION.GetString(IDS_HOW_TO_DISPLAY_HELP));
		printf(strTemp, PROGRAM_NAME, PRG_ARG_HELP);
		exit(0);
	}

	// --- Display device list ---
	if (bDisplayDeviceList)
	{
		CDManager	 detectManager;
		detectManager.Initialize(NULL);
		detectManager.ScanDevices();

		printf("%s %s\n\n", PROGRAM_NAME, INFO_CONTROL_PANEL);
		for(i=0;i<MAX_DEVICES;++i)
		{
			if (g_DEVICE_INFO[i].bPresent)
			{
				strcpy(strTemp, GetDeviceName(g_DEVICE_INFO[i].ulProductCode));
				j = 25-strlen(strTemp);
				while(j-->0) 
					strcat(strTemp, " ");
				printf("%s - ", strTemp);
				printf("%s %s\n", g_LOCALIZATION.GetString(IDS_PHYSICAL_LOCATION),
					g_DEVICE_INFO[i].szLocationID);
			}
		}
		exit(0);
	}

	// --- Detect if another instance is running, if so exit ---
	if (!bAllowMultipleInstance && processCount()>1)
	{
		sprintf(strTemp, "%s\n", g_LOCALIZATION.GetString(IDS_WARNING_PROCESS_ALREADY_RUN));
		printf(strTemp, PROGRAM_NAME);
		exit(0);
	}

	DEBUG_PRINT("Finalizing initialization...");
	g_pDeviceSelect = pDevice;

#if wxUSE_LIBPNG
    	wxImage::AddHandler( new wxPNGHandler );
#endif
	wxIcon icon(wxICON(hdjcpl));

#ifdef WIN32
    	m_frame = new MyDlg( _T(APP_NAME), wxPoint(200,250), wxSize(546+6,347+32/*538+7,400+32*/) );
#else
    	m_frame = new MyDlg( _T(APP_NAME), wxPoint(200,250), wxSize(546,347/*538+7,400+32*/) );
#endif

	// --- getenv("LANG") return a code which look like "en_CA.UTF-8" ---
	// The first 2 characters define the language:
	//
	// en: english
	// fr: french
	// it: italian
	// de: German
	// nl: Dutch
	// es: spannish
	// pt: portuguese
	//
	m_frame->ShowModal();
	m_frame->Destroy();

	DEBUG_PRINT("About to exit!");
	return TRUE;
} 


BEGIN_EVENT_TABLE(MyDlg, wxDialog)
    	EVT_PAINT (MyDlg::OnPaint)
	EVT_MOTION(MyDlg::OnMouseMotion)
	EVT_RIGHT_UP(MyDlg::OnMouseRightUp)
	EVT_LEFT_UP(MyDlg::OnMouseLeftUp)
	EVT_LEFT_DOWN(MyDlg::OnMouseLeftDown)
	EVT_LEFT_DCLICK(MyDlg::OnMouseDClick)
    	EVT_TIMER(TIMER_ID, MyDlg::OnTimer)
	EVT_TIMER(TIMER_CHECK_DEV_CHANGE_ID, MyDlg::OnCheckDevChange)
	EVT_SIZE(MyDlg::OnSizing)
	EVT_MENU(IDM_DEVICE_INDEX_0, MyDlg::OnIDM_DeviceIndex0)
	EVT_MENU(IDM_DEVICE_INDEX_1, MyDlg::OnIDM_DeviceIndex1)
	EVT_MENU(IDM_DEVICE_INDEX_2, MyDlg::OnIDM_DeviceIndex2)
	EVT_MENU(IDM_DEVICE_INDEX_3, MyDlg::OnIDM_DeviceIndex3)
	EVT_MENU(IDM_DEVICE_INDEX_4, MyDlg::OnIDM_DeviceIndex4)
	EVT_MENU(IDM_DEVICE_INDEX_5, MyDlg::OnIDM_DeviceIndex5)
	EVT_MENU(IDM_DEVICE_INDEX_6, MyDlg::OnIDM_DeviceIndex6)
	EVT_MENU(IDM_DEVICE_INDEX_7, MyDlg::OnIDM_DeviceIndex7)
	EVT_MENU(IDM_DEVICE_INDEX_8, MyDlg::OnIDM_DeviceIndex8)
	EVT_MENU(IDM_DEVICE_INDEX_9, MyDlg::OnIDM_DeviceIndex9)
	EVT_MENU(IDM_DEVICE_INDEX_10, MyDlg::OnIDM_DeviceIndex10)
	EVT_MENU(IDM_DEVICE_INDEX_11, MyDlg::OnIDM_DeviceIndex11)
	EVT_MENU(IDM_DEVICE_INDEX_12, MyDlg::OnIDM_DeviceIndex12)
	EVT_MENU(IDM_DEVICE_INDEX_13, MyDlg::OnIDM_DeviceIndex13)
	EVT_MENU(IDM_DEVICE_INDEX_14, MyDlg::OnIDM_DeviceIndex14)
	EVT_MENU(IDM_DEVICE_INDEX_15, MyDlg::OnIDM_DeviceIndex15)
END_EVENT_TABLE()

IMPLEMENT_APP(MyApp)


// ---------------------------------------------------------------------------
// Function: MyDlg
//
// Main program dialog constructor.
// ---------------------------------------------------------------------------
MyDlg::MyDlg(const wxString& title, const wxPoint& pos, const wxSize& size)
: wxDialog((wxWindow *)NULL, -1, title, pos, size, wxBORDER_DEFAULT|wxMINIMIZE_BOX|wxRESIZE_BORDER|wxDEFAULT_DIALOG_STYLE), 
m_timer(this, TIMER_ID), m_timerDevChange(this, TIMER_CHECK_DEV_CHANGE_ID)
{
	int i;
	BrInitError error; 
	char *strDataDir;
	char szFilename[512];
	char strTemp[260];
	struct stat stat_buf;

	// --- Set main dialog properties ---
	DEBUG_PRINT("Set main dialog properties...");
	SetIcon(hdjcpl_xpm);	// Single XPM file with 1-bit Alpha.
	m_bCloseDown = FALSE;
	m_bCloseOver = FALSE;
	m_pDialog = NULL;
	
	// --- Determine where the external resources are located ---
	// we use binreloc from AutoPackage.org
	// We first look under ../data which is where the files are if we are
	// running the application in the source tree.
	// Otherwise, we look to ../share/hercules-hdjcpl where the PNG files
	// are copied during the installation (packages and make install).
	DEBUG_PRINT("Determine where the external resources files are located...");
	if (br_init (&error) == 0 && error != BR_INIT_ERROR_DISABLED) { 
		DEBUG_PRINT("Warning: BinReloc failed to initialize (error code %d)", error); 
		DEBUG_PRINT("Will fallback to hardcoded default path.");
	} 
	
	strDataDir = br_find_exe_dir(".");
	strcpy(m_szBasePath, strDataDir);
	strcat(m_szBasePath, "/../data/");
	free(strDataDir);
	strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-djc.png");
	if (stat(szFilename, &stat_buf)<0)
	{	
		strDataDir = br_find_data_dir("../share");
		strcpy(m_szBasePath, strDataDir);
		free(strDataDir);
		strcat(m_szBasePath, "/hercules-hdjcpl/");
	}
	DEBUG_PRINT("External files path to be use: %s", m_szBasePath);

	// --- Continue intialization ---
	SetBackgroundStyle(wxBG_STYLE_CUSTOM);
	m_DManager.Initialize(this);
	g_RescanDeviceList = 0;
	m_DManager.ScanDevices();

	// --- Determine which device to select ---
	DEBUG_PRINT("Determine which device to select...");
	m_nDeviceIndex = -1;
	if (g_pDeviceSelect!=NULL)
	{
		for(i=0;m_nDeviceIndex==-1 && i<MAX_DEVICES;++i)
		{
			if (g_DEVICE_INFO[i].bPresent && strcasecmp(g_pDeviceSelect, g_DEVICE_INFO[i].szLocationID)==0)
				m_nDeviceIndex = i;
		}

		if (m_nDeviceIndex<0)
		{
			sprintf(strTemp, "%s\n", g_LOCALIZATION.GetString(IDS_WARNING_DEVICE_NOT_FOUND));
			printf(strTemp, g_pDeviceSelect);
		}
	}
	
	for(i=0;m_nDeviceIndex==-1 && i<MAX_DEVICES;++i)
	{
		if (g_DEVICE_INFO[i].bPresent)
			m_nDeviceIndex = i;
	}
	// --- Select the device which has been selected ---
	DEBUG_PRINT("At initialization, we use device index = %d.", m_nDeviceIndex);
	ChangeDevice( m_nDeviceIndex );

	// --- Start application Timers ---
	m_timerDevChange.Start(TIMER_CHECK_DEV_CHANGE_DELAY, wxTIMER_ONE_SHOT);    // 1/10 second interval
	DEBUG_PRINT("Timer started, main dialog box constructor completed.");
}

// ---------------------------------------------------------------------------
// Function: ~MyDlg
//
// Main program dialog destructor.
// ---------------------------------------------------------------------------
MyDlg::~MyDlg()
{
	if (m_pDialog!=NULL)
	{
		m_dashboard.Destroy();
		m_dashboardAlpha.Destroy();
		SAFE_DELETE(m_pBmpDashboard);
		SAFE_DELETE(m_pBmpDashboardAlpha);
		SAFE_DELETE(m_pDialog);
	}
}

// ---------------------------------------------------------------------------
// Function: OnCheckDevChange
//
// Main program dialog destructor.
// ---------------------------------------------------------------------------
void MyDlg::OnCheckDevChange(wxTimerEvent& event)
{
	int i;
	bool bRefresh = false;

	// --- ALSA Mixer ---
	if (m_nDeviceIndex>=0)
	{
		// --- Check device state change ---
		if (g_DEVICE_INFO[m_nDeviceIndex].bStateChange)
		{
			g_DEVICE_INFO[m_nDeviceIndex].bStateChange = false;
			m_pDialog->InitializeDlgState(&g_DEVICE_INFO[m_nDeviceIndex], &m_DManager);
			bRefresh |= true;
		}

		if (g_DEVICE_INFO[m_nDeviceIndex].bALSA_Init)
		{
			bRefresh |= m_pDialog->ApplyVolume(&g_DEVICE_INFO[m_nDeviceIndex]);

			m_DManager.UpdateAlsa(&g_DEVICE_INFO[m_nDeviceIndex]);
			if (g_DEVICE_INFO[m_nDeviceIndex].bALSA_Refresh)
			{
				g_DEVICE_INFO[m_nDeviceIndex].bALSA_Refresh = false;
				m_pDialog->InitializeMixerState(&g_DEVICE_INFO[m_nDeviceIndex], &m_DManager);
				bRefresh = true;
			}
		}
	}

	if (g_RescanDeviceList==0) {
		if (bRefresh)
			Refresh();
		m_timerDevChange.Start(TIMER_CHECK_DEV_CHANGE_DELAY, wxTIMER_ONE_SHOT);    // 1/10 second interval
		return;
	}

	DEBUG_PRINT("Re-scanning device list is required.");
	g_RescanDeviceList = 0;
	m_DManager.ScanDevices();

	// --- Set current device index ---
	if (m_nDeviceIndex!=-1)	{ 
		if (!g_DEVICE_INFO[m_nDeviceIndex].bPresent)
		{
			m_nDeviceIndex = -1;
			m_DManager.SelectDevice( m_nDeviceIndex );
		}
	}

	for(i=0;m_nDeviceIndex==-1 && i<MAX_DEVICES;++i)
	{
		if (g_DEVICE_INFO[i].bPresent)
			m_nDeviceIndex = i;
	}

	DEBUG_PRINT("Change device to index %d.", m_nDeviceIndex);
	ChangeDevice( m_nDeviceIndex );
	m_timerDevChange.Start(TIMER_CHECK_DEV_CHANGE_DELAY, wxTIMER_ONE_SHOT);    // 1/10 second interval
}

// ---------------------------------------------------------------------------
// Function: OnTimer
// ---------------------------------------------------------------------------
void MyDlg::OnTimer(wxTimerEvent& event)
{
	Refresh();
}

// ---------------------------------------------------------------------------
// Function: ChangeDevice
//
// Change the dialog box (look and class used) related to device selected.
// ---------------------------------------------------------------------------
void MyDlg::ChangeDevice( int nIndex )
{
	int nWidth, nHeight, nPosX, nPosY;
	char szFilename[512];
	bool bPngLoaded = false;

	m_nDeviceIndex=nIndex;
	if (m_pDialog!=NULL)
	{
		DEBUG_PRINT("Deleting previous resources (dashboard and dialog).");
		m_dashboard.Destroy();
		m_dashboardAlpha.Destroy();
		SAFE_DELETE(m_pBmpDashboard);
		SAFE_DELETE(m_pBmpDashboardAlpha);
		SAFE_DELETE(m_pDialog);
		DEBUG_PRINT("Resources deleted.");
	}

	if (m_nDeviceIndex>=0)
	{
		switch(g_DEVICE_INFO[m_nDeviceIndex].ulProductCode)
		{
			case DJCONSOLE_PRODUCT_CODE:
				DEBUG_PRINT("Upcoming resources to be loaded: DJ Console.");
				strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-djc.png");
				m_dashboard = wxImage(szFilename, wxBITMAP_TYPE_PNG);
				strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-djc-alpha.png");
				m_dashboardAlpha = wxImage(szFilename, wxBITMAP_TYPE_PNG);
				bPngLoaded = true;
				break;
			case DJCONTROLLER_PRODUCT_CODE:
				DEBUG_PRINT("Upcoming resources to be loaded: DJ Control MP3.");
				strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-mp3.png");
				m_dashboard = wxImage(szFilename, wxBITMAP_TYPE_PNG);
				strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-mp3-alpha.png");
				m_dashboardAlpha = wxImage(szFilename, wxBITMAP_TYPE_PNG);
				bPngLoaded = true;
				break;
			case DJCONSOLE2_PRODUCT_CODE:
				DEBUG_PRINT("Upcoming resources to be loaded: DJ Console Mk2.");
				strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-mk2.png");
				m_dashboard = wxImage(szFilename, wxBITMAP_TYPE_PNG);
				strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-mk2-alpha.png");
				m_dashboardAlpha = wxImage(szFilename, wxBITMAP_TYPE_PNG);
				bPngLoaded = true;
				break;
			case DJCONSOLERMX_PRODUCT_CODE:
				DEBUG_PRINT("Upcoming resources to be loaded: DJ Console Rmx.");
				strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-rmx.png");
				m_dashboard = wxImage(szFilename, wxBITMAP_TYPE_PNG);
				strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-rmx-alpha.png");
				m_dashboardAlpha = wxImage(szFilename, wxBITMAP_TYPE_PNG);
				bPngLoaded = true;
				break;
			case DJCONTROLSTEEL_PRODUCT_CODE:
				DEBUG_PRINT("Upcoming resources to be loaded: DJ Control Steel.");
				strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-steel.png");
				m_dashboard = wxImage(szFilename, wxBITMAP_TYPE_PNG);
				strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-steel-alpha.png");
				m_dashboardAlpha = wxImage(szFilename, wxBITMAP_TYPE_PNG);
				bPngLoaded = true;
				break;
		}
	}

	if (!bPngLoaded)
	{
		DEBUG_PRINT("Upcoming resources to be loaded: null (no device dialog).");
		strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-null.png");
		m_dashboard = wxImage(szFilename, wxBITMAP_TYPE_PNG);
		strcpy(szFilename, m_szBasePath); strcat(szFilename, "bg-null-alpha.png");
		m_dashboardAlpha = wxImage(szFilename, wxBITMAP_TYPE_PNG);
	}

	DEBUG_PRINT("Loading resources...");
	m_pBmpDashboard = new wxBitmap(m_dashboard);
	m_pBmpDashboardAlpha = new wxBitmap(m_dashboardAlpha);
	m_memDCDashboard.SelectObject(*m_pBmpDashboard);
	m_memDCDashboardAlpha.SelectObject(*m_pBmpDashboardAlpha);

	DEBUG_PRINT("Create dialog object...");
	if (bPngLoaded)
	{
		switch(g_DEVICE_INFO[m_nDeviceIndex].ulProductCode)
		{
		case DJCONSOLE_PRODUCT_CODE:
			m_pDialog = (CGDialog *) new CDevDjcDlg(&m_memDCDashboard, &m_memDCDashboardAlpha);
			break;
		case DJCONTROLLER_PRODUCT_CODE:
			m_pDialog = (CGDialog *) new CDevMp3Dlg(&m_memDCDashboard, &m_memDCDashboardAlpha);
			break;
		case DJCONSOLE2_PRODUCT_CODE:
			m_pDialog = (CGDialog *) new CDevMk2Dlg(&m_memDCDashboard, &m_memDCDashboardAlpha);
			break;
		case DJCONSOLERMX_PRODUCT_CODE:
			m_pDialog = (CGDialog *) new CDevRmxDlg(&m_memDCDashboard, &m_memDCDashboardAlpha);
			break;
		case DJCONTROLSTEEL_PRODUCT_CODE:
			m_pDialog = (CGDialog *) new CDevSteelDlg(&m_memDCDashboard, &m_memDCDashboardAlpha);
			break;
		}
		m_pDialog->InitializeDlgState(&g_DEVICE_INFO[m_nDeviceIndex], &m_DManager);
	}
	else
		m_pDialog = (CGDialog *) new CDevNullDlg(&m_memDCDashboard, &m_memDCDashboardAlpha);

	DEBUG_PRINT("Select device index %d.", m_nDeviceIndex);
	m_DManager.SelectDevice( m_nDeviceIndex );

	// --- Change Dialog Size ---
	GetPosition(&nPosX, &nPosY);
	m_pDialog->GetSize(&nWidth, &nHeight);

	wxSize sizeA(nWidth, nHeight);
	wxSize minSize(1,1);
	wxSize maxSize(800,480);

	// Set constrain to something which allow all size we use.
	// we have choosen 800,480 for the upper limit (which is the one we know on Guillemot UMPC).
	SetMaxSize(maxSize);
	SetMinSize(minSize);

	SetSize(nPosX, nPosY, nWidth, nHeight);

	// Set contrain to the same as the size we just set (this prevent the user to resize the dialog box).
	SetMaxSize(sizeA);
	SetMinSize(sizeA);

	DEBUG_PRINT("Dialog size and constrain has been apply.");
	Refresh();
}

// ---------------------------------------------------------------------------
// Function: OnSizing
//
// Adjust the size to what it should be (if needed).
// This function should not be call, as we set the min and max size to the
// current size. With wxGTK, this is not a problem, but prior to that we try
// with wxX11 (wxWidgets version 2.8.7) under Ubuntu 8.04 and the result was
// that we were always able to resize the dialog box (however with this code,
// once the user release the sizing, the size was restore to normal).
// ---------------------------------------------------------------------------
void MyDlg::OnSizing( wxSizeEvent &event )
{
	int nPosX, nPosY, nWidth, nHeight;
	int cx, cy;
		
	GetPosition(&nPosX, &nPosY);
	GetSize(&cx, &cy);
	m_pDialog->GetSize(&nWidth, &nHeight);

	if (cx!=nWidth || cy!=nHeight)
	{
		SetSize(nPosX, nPosY, nWidth, nHeight);
		wxSize sizeA(nWidth, nHeight);
		Refresh();
	}
}

// ---------------------------------------------------------------------------
// Function: OnMouseMotion
// ---------------------------------------------------------------------------
void MyDlg::OnMouseMotion(wxMouseEvent &event)
{
	wxClientDC	dc(this);
	int		nButton;
	int		nDelay;
	wxRect		rect;
	REFRESH_RECT 	refreshRect;

	nButton = (event.LeftIsDown()) ? GMOUSE_LEFT_DOWN : 0;
	nButton|= (event.RightIsDown()) ? GMOUSE_RIGHT_DOWN : 0;
	if (m_pDialog->MouseEvent(event.GetX(), event.GetY(), nButton, &refreshRect)) {
		m_pDialog->Paint(&dc, &refreshRect);
		nDelay = m_pDialog->GetDelayForNextToolTipsEvent();

		if (nDelay>0) {
			m_timer.Start(nDelay);    // 1/10 second interval
		}
	}

	return;
}

// ---------------------------------------------------------------------------
// Function: OnMouseLeftDown
// ---------------------------------------------------------------------------
void MyDlg::OnMouseLeftDown(wxMouseEvent &event)
{
	wxClientDC	dc(this);
	int		nButton;
	REFRESH_RECT 	refreshRect;

	nButton = (event.LeftIsDown()) ? GMOUSE_LEFT_DOWN : 0;
	nButton|= (event.RightIsDown()) ? GMOUSE_RIGHT_DOWN : 0;
	if (m_pDialog->MouseEvent(event.GetX(), event.GetY(), nButton, &refreshRect))
		m_pDialog->Paint(&dc, &refreshRect);
}

// ---------------------------------------------------------------------------
// Function: OnMouseLeftUp
// ---------------------------------------------------------------------------
void MyDlg::OnMouseLeftUp (wxMouseEvent &event)
{
	wxClientDC	dc(this);
	int		nButton;
	REFRESH_RECT	refreshRect;

	nButton = (event.LeftIsDown()) ? GMOUSE_LEFT_DOWN : 0;
	nButton|= (event.RightIsDown()) ? GMOUSE_RIGHT_DOWN : 0;
	if (m_pDialog->MouseEvent(event.GetX(), event.GetY(), nButton, &refreshRect))
		m_pDialog->Paint(&dc, &refreshRect);
}

// ---------------------------------------------------------------------------
// Function: GetDeviceName
//
// Return the device name in char* from the device type code.
// ---------------------------------------------------------------------------
#define MAX_DEV_PROD_CODE	6
char *GetDeviceName( int nDevType )
{
	static char szName[MAX_DEV_PROD_CODE][260] = {
		"Hercules DJ Unknown",
		"Hercules DJ Console",
		"Hercules DJ Console Mk2",
		"Hercules DJ Control MP3",
		"Hercules DJ Console Rmx",
		"Hercules DJ Control Steel"
	};
	int nDevIndex = nDevType;

	if (nDevIndex>(int) (sizeof(szName)/sizeof(szName[0])) )
		nDevIndex = 0;

	return szName[nDevIndex];
}

// ---------------------------------------------------------------------------
// AddDeviceNameToMenu
//
// When the user right click on the dialog box, we display a context menu
// showing all devices available. This function, add an item base on its
// informations.
// ---------------------------------------------------------------------------
bool AddDeviceNameToMenu( int nMenuCode, int nDevType, int *pnDevCount, wxMenu *pMenu )
{
	char	szDevName[260], szExtDevName[260];
	int 	nDevIndex = nDevType;
	
	if (pnDevCount==NULL || pMenu==NULL)
		return false;

	if (nDevIndex>=MAX_DEV_PROD_CODE)
		nDevIndex = 0;
	strcpy(szDevName, GetDeviceName(nDevType));
	sprintf(szExtDevName, " (%d)", ++pnDevCount[nDevIndex]);
	if (pnDevCount[nDevIndex]>1) 
		strcat(szDevName, szExtDevName);

	pMenu->Append(nMenuCode, szDevName);
	return true;
}

// ---------------------------------------------------------------------------
// OnMouseRightUp
//
// When the user right click on the dialog box, we display a context menu
// showing all devices available.
// ---------------------------------------------------------------------------
void MyDlg::OnMouseRightUp(wxMouseEvent &event)
{
	wxMenu	menu;
	int	i, nCount, nCountTable[MAX_DEV_PROD_CODE];

	nCount = 0;
	for(i=0;i<(int) (sizeof(nCountTable)/sizeof(nCountTable[0]));++i)
		nCountTable[i] = 0;

	for(i=0;i<MAX_DEVICES;++i)
	{
		if (g_DEVICE_INFO[i].bPresent)
		{
			if (AddDeviceNameToMenu(i, g_DEVICE_INFO[i].ulProductCode, nCountTable, &menu))
				++nCount;
		}
	}

	if (nCount>0)
	{
		DEBUG_PRINT("Pop-Up contextual menu with device present.");
		PopupMenu(&menu, event.GetX(), event.GetY());
		SetTitle(APP_NAME);
	}
}

// ---------------------------------------------------------------------------
// OnMouseDClick
// ---------------------------------------------------------------------------
void MyDlg::OnMouseDClick(wxMouseEvent &event)
{
	wxClientDC	dc(this);
	int		nButton;
	REFRESH_RECT 	refreshRect;

	nButton = (event.LeftIsDown()) ? GMOUSE_LEFT_DOWN : 0;
	nButton|= (event.RightIsDown()) ? GMOUSE_RIGHT_DOWN : 0;
	if (m_pDialog->MouseEvent(event.GetX(), event.GetY(), nButton|GMOUSE_LEFT_DCLICK, &refreshRect))
		m_pDialog->Paint(&dc, &refreshRect);
}

// ---------------------------------------------------------------------------
// OnPaint
// ---------------------------------------------------------------------------
void MyDlg::OnPaint (wxPaintEvent &event)
{
	int		nDelay;
	wxPaintDC	dcP(this);
	REFRESH_RECT 	refreshRect;

	refreshRect.x = 0; refreshRect.y = 0;
	m_pDialog->GetSize(&refreshRect.w, &refreshRect.h);

	m_pDialog->Paint(&dcP, &refreshRect);
	nDelay = m_pDialog->GetDelayForNextToolTipsEvent();

	if (nDelay>0) {
	    m_timer.Start(nDelay);    // 1/10 second interval
	}
}

// ---------------------------------------------------------------------------
// Destroy
//
// By selecting -1 for the device, we ensure that the current device is
// release (closing handle on it and freeing the resources).
// ---------------------------------------------------------------------------
bool MyDlg::Destroy( void )
{
	DEBUG_PRINT("Destroying main dialog object.");
	m_DManager.SelectDevice(-1);
	m_DManager.Uninitialize();
	return wxDialog::Destroy();
}


